#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022 - 2023

import json
import logging
import time
import threading

from idds.common.constants import (ProcessingStatus,
                                   CollectionStatus,
                                   ContentStatus, ContentType,
                                   ContentRelationType,
                                   WorkStatus,
                                   TransformType,
                                   TransformType2MessageTypeMap,
                                   MessageType, MessageTypeStr,
                                   MessageStatus, MessageSource,
                                   MessageDestination)
from idds.common.utils import setup_logging
from idds.core import (transforms as core_transforms,
                       processings as core_processings,
                       catalog as core_catalog)
from idds.agents.common.cache.redis import get_redis_cache


setup_logging(__name__)


def get_logger(logger=None):
    if logger:
        return logger
    logger = logging.getLogger(__name__)
    return logger


def get_new_content(request_id, transform_id, workload_id, map_id, input_content, content_relation_type=ContentRelationType.Input):
    content = {'transform_id': transform_id,
               'coll_id': input_content['coll_id'],
               'request_id': request_id,
               'workload_id': workload_id,
               'map_id': map_id,
               'scope': input_content['scope'],
               'name': input_content['name'],
               'min_id': input_content['min_id'] if 'min_id' in input_content else 0,
               'max_id': input_content['max_id'] if 'max_id' in input_content else 0,
               'status': input_content['status'] if 'status' in input_content and input_content['status'] is not None else ContentStatus.New,
               'substatus': input_content['substatus'] if 'substatus' in input_content and input_content['substatus'] is not None else ContentStatus.New,
               'path': input_content['path'] if 'path' in input_content else None,
               'content_type': input_content['content_type'] if 'content_type' in input_content else ContentType.File,
               'content_relation_type': content_relation_type,
               'bytes': input_content['bytes'],
               'adler32': input_content['adler32'],
               'content_metadata': input_content['content_metadata']}
    if content['min_id'] is None:
        content['min_id'] = 0
    if content['max_id'] is None:
        content['max_id'] = 0
    return content


def is_process_terminated(processing_status):
    if processing_status in [ProcessingStatus.Finished, ProcessingStatus.Failed,
                             ProcessingStatus.SubFinished, ProcessingStatus.Cancelled,
                             ProcessingStatus.Suspended, ProcessingStatus.Expired,
                             ProcessingStatus.Broken, ProcessingStatus.FinishedOnStep,
                             ProcessingStatus.FinishedOnExec, ProcessingStatus.FinishedTerm]:
        return True
    return False


def is_process_finished(processing_status):
    if processing_status in [ProcessingStatus.Finished]:
        return True
    return False


def is_all_contents_available(contents):
    for content in contents:
        if type(content) is dict:
            if content['substatus'] not in [ContentStatus.Available, ContentStatus.FakeAvailable]:
                return False
        else:
            # list of content_id, status,
            if content[1] not in [ContentStatus.Available, ContentStatus.FakeAvailable]:
                return False
    return True


def is_all_contents_terminated(contents):
    for content in contents:
        if type(content) is dict:
            if content['substatus'] not in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                            ContentStatus.FinalFailed, ContentStatus.Missing]:
                return False
        else:
            if content[1] not in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                  ContentStatus.FinalFailed, ContentStatus.Missing]:
                return False
    return True


def is_input_dependency_terminated(input_dependency):
    if type(input_dependency) is dict:
        if input_dependency['substatus'] in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                             ContentStatus.FinalFailed, ContentStatus.Missing]:
            return True
    else:
        if input_dependency[1] in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                   ContentStatus.FinalFailed, ContentStatus.Missing]:
            return True
    return False


def is_all_contents_terminated_but_not_available(inputs):
    all_contents_available = True
    for content in inputs:
        if type(content) is dict:
            if content['substatus'] not in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                            ContentStatus.FinalFailed, ContentStatus.Missing]:
                return False
            if content['substatus'] not in [ContentStatus.Available]:
                all_contents_available = False
        else:
            if content[1] not in [ContentStatus.Available, ContentStatus.FakeAvailable,
                                  ContentStatus.FinalFailed, ContentStatus.Missing]:
                return False
            if content[1] not in [ContentStatus.Available]:
                all_contents_available = False
    if all_contents_available:
        return False
    return True


def is_all_contents_available_with_status_map(inputs_dependency, content_status_map):
    for content_id in inputs_dependency:
        status = content_status_map[str(content_id)]
        if status not in [ContentStatus.Available, ContentStatus.FakeAvailable]:
            return False
    return True


def is_all_contents_terminated_with_status_map(input_dependency, content_status_map):
    for content_id in input_dependency:
        status = content_status_map[str(content_id)]
        if status not in [ContentStatus.Available, ContentStatus.FakeAvailable,
                          ContentStatus.FinalFailed, ContentStatus.Missing]:
            return False
    return True


def get_collection_ids(collections):
    coll_ids = [coll.coll_id for coll in collections]
    return coll_ids


def get_input_output_maps(transform_id, work):
    # link collections
    input_collections = work.get_input_collections()
    output_collections = work.get_output_collections()
    log_collections = work.get_log_collections()

    # for coll in input_collections + output_collections + log_collections:
    #     coll_model = core_catalog.get_collection(coll_id=coll.coll_id)
    #     coll.collection = coll_model

    input_coll_ids = get_collection_ids(input_collections)
    output_coll_ids = get_collection_ids(output_collections)
    log_coll_ids = get_collection_ids(log_collections)

    mapped_input_output_maps = core_transforms.get_transform_input_output_maps(transform_id,
                                                                               input_coll_ids=input_coll_ids,
                                                                               output_coll_ids=output_coll_ids,
                                                                               log_coll_ids=log_coll_ids)

    # work_name_to_coll_map = core_transforms.get_work_name_to_coll_map(request_id=transform['request_id'])
    # work.set_work_name_to_coll_map(work_name_to_coll_map)

    # new_input_output_maps = work.get_new_input_output_maps(mapped_input_output_maps)
    return mapped_input_output_maps


def get_ext_contents(transform_id, work):
    contents_ids = core_catalog.get_contents_ext_ids(transform_id=transform_id)
    return contents_ids


def resolve_input_dependency_id(request_id, new_input_dep_coll_ids, new_input_dependency_contents, logger=None, log_prefix=''):
    logger = get_logger(logger)

    logger.info(log_prefix + "resolve_input_dependency_id: new_input_dep_coll_ids: %s" % (str(new_input_dep_coll_ids)))
    logger.info(log_prefix + "resolve_input_dependency_id: len(new_input_dependency_contents): %s" % len(new_input_dependency_contents))

    if new_input_dep_coll_ids:
        contents = core_catalog.get_contents_by_coll_id_status(coll_id=new_input_dep_coll_ids)
        content_name_id_map = {}
        for content in contents:
            if content['coll_id'] not in content_name_id_map:
                content_name_id_map[content['coll_id']] = {}
            content_name_id_map[content['coll_id']][content['name']] = content['content_id']
        for content in new_input_dependency_contents:
            if content['coll_id'] in content_name_id_map and content['name'] in content_name_id_map[content['coll_id']]:
                content_dep_id = content_name_id_map[content['coll_id']][content['name']]
                content['content_dep_id'] = content_dep_id
    return new_input_dependency_contents


def get_new_contents(request_id, transform_id, workload_id, new_input_output_maps, logger=None, log_prefix=''):
    logger = get_logger(logger)

    logger.debug(log_prefix + "get_new_contents")
    new_input_contents, new_output_contents, new_log_contents = [], [], []
    new_input_dependency_contents = []
    new_input_dep_coll_ids = []
    for map_id in new_input_output_maps:
        inputs = new_input_output_maps[map_id]['inputs'] if 'inputs' in new_input_output_maps[map_id] else []
        inputs_dependency = new_input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in new_input_output_maps[map_id] else []
        outputs = new_input_output_maps[map_id]['outputs'] if 'outputs' in new_input_output_maps[map_id] else []
        logs = new_input_output_maps[map_id]['logs'] if 'logs' in new_input_output_maps[map_id] else []

        for input_content in inputs:
            content = get_new_content(request_id, transform_id, workload_id, map_id, input_content, content_relation_type=ContentRelationType.Input)
            new_input_contents.append(content)
        for input_content in inputs_dependency:
            content = get_new_content(request_id, transform_id, workload_id, map_id, input_content, content_relation_type=ContentRelationType.InputDependency)
            new_input_dependency_contents.append(content)
            if content['coll_id'] not in new_input_dep_coll_ids:
                new_input_dep_coll_ids.append(content['coll_id'])
        for output_content in outputs:
            content = get_new_content(request_id, transform_id, workload_id, map_id, output_content, content_relation_type=ContentRelationType.Output)
            new_output_contents.append(content)
        for log_content in logs:
            content = get_new_content(request_id, transform_id, workload_id, map_id, log_content, content_relation_type=ContentRelationType.Log)
            new_log_contents.append(content)

    # new_input_dependency_contents = resolve_input_dependency_id(request_id, new_input_dep_coll_ids, new_input_dependency_contents, logger=logger, log_prefix=log_prefix)
    return new_input_contents, new_output_contents, new_log_contents, new_input_dependency_contents


def get_update_content(content):
    updated_content = {'content_id': content['content_id'],
                       'status': content['substatus'],
                       'substatus': content['substatus']}
    content['status'] = content['substatus']
    return updated_content, content


def get_update_contents(request_id, transform_id, workload_id, input_output_maps):
    updated_contents = []
    updated_input_contents_full, updated_output_contents_full = [], []

    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        # logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        content_update_status = None
        if is_all_contents_available(inputs_dependency):
            # logger.debug("all input dependency available: %s, inputs: %s" % (str(inputs_dependency), str(inputs)))
            content_update_status = ContentStatus.Available
        elif is_all_contents_terminated(inputs_dependency):
            # logger.debug("all input dependency terminated: %s, inputs: %s, outputs: %s" % (str(inputs_dependency), str(inputs), str(outputs)))
            content_update_status = ContentStatus.Missing

        if content_update_status:
            for content in inputs:
                content['substatus'] = content_update_status
                if content['status'] != content['substatus']:
                    updated_content, content = get_update_content(content)
                    updated_contents.append(updated_content)
                    updated_input_contents_full.append(content)
        if content_update_status in [ContentStatus.Missing]:
            for content in outputs:
                content['substatus'] = content_update_status
                if content['status'] != content['substatus']:
                    updated_content, content = get_update_content(content)
                    updated_contents.append(updated_content)
                    updated_output_contents_full.append(content)

        for content in outputs:
            if content['status'] != content['substatus']:
                updated_content, content = get_update_content(content)
                updated_contents.append(updated_content)
                updated_output_contents_full.append(content)
        return updated_contents, updated_input_contents_full, updated_output_contents_full


def get_message_type(work_type, input_type='file'):
    work_type_value = str(work_type.value)
    if work_type_value not in TransformType2MessageTypeMap:
        return TransformType2MessageTypeMap['0'][input_type]
    else:
        return TransformType2MessageTypeMap[work_type_value][input_type]


def generate_file_messages(request_id, transform_id, workload_id, work, files, relation_type):
    if work:
        work_type = work.get_work_type()
    else:
        work_type = TransformType.Processing

    i_msg_type, i_msg_type_str = get_message_type(work_type, input_type='file')
    files_message = []
    for file in files:
        file_status = file['substatus'].name
        if file['substatus'] == ContentStatus.FakeAvailable:
            file_status = ContentStatus.Available.name
        file_message = {'scope': file['scope'],
                        'name': file['name'],
                        'path': file['path'],
                        'status': file_status}
        files_message.append(file_message)
    msg_content = {'msg_type': i_msg_type_str.value,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'relation_type': relation_type,
                   'files': files_message}
    num_msg_content = len(files_message)
    return i_msg_type, msg_content, num_msg_content


def generate_content_ext_messages(request_id, transform_id, workload_id, work, files, relation_type, input_output_maps):
    i_msg_type = MessageType.ContentExt
    i_msg_type_str = MessageTypeStr.ContentExt

    output_contents = []
    for map_id in input_output_maps:
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        for content in outputs:
            # content_map[content['content_id']] = content
            output_contents += outputs

    files_message = core_catalog.combine_contents_ext(output_contents, files, with_status_name=True)
    msg_content = {'msg_type': i_msg_type_str.value,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'relation_type': relation_type,
                   'files': files_message}
    num_msg_content = len(files_message)
    return i_msg_type, msg_content, num_msg_content


def generate_collection_messages(request_id, transform_id, workload_id, work, collection, relation_type):
    coll_name = collection.name
    if coll_name.endswith(".idds.stagein"):
        coll_name = coll_name.replace(".idds.stagein", "")

    i_msg_type, i_msg_type_str = get_message_type(work.get_work_type(), input_type='collection')
    msg_content = {'msg_type': i_msg_type_str.value,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'relation_type': relation_type,
                   'collections': [{'scope': collection.scope,
                                    'name': coll_name,
                                    'status': collection.status.name}],
                   'output': work.get_output_data(),
                   'error': work.get_terminated_msg()}
    num_msg_content = 1
    return i_msg_type, msg_content, num_msg_content


def generate_work_messages(request_id, transform_id, workload_id, work, relation_type):
    i_msg_type, i_msg_type_str = get_message_type(work.get_work_type(), input_type='work')
    msg_content = {'msg_type': i_msg_type_str.value,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'relation_type': relation_type,
                   'status': work.get_status().name,
                   'output': work.get_output_data(),
                   'error': work.get_terminated_msg()}
    num_msg_content = 1
    return i_msg_type, msg_content, num_msg_content


def generate_messages(request_id, transform_id, workload_id, work, msg_type='file', files=[], relation_type='input', input_output_maps=None):
    if msg_type == 'file':
        i_msg_type, msg_content, num_msg_content = generate_file_messages(request_id, transform_id, workload_id, work, files=files, relation_type=relation_type)
        msg = {'msg_type': i_msg_type,
               'status': MessageStatus.New,
               'source': MessageSource.Carrier,
               'destination': MessageDestination.Outside,
               'request_id': request_id,
               'workload_id': workload_id,
               'transform_id': transform_id,
               'num_contents': num_msg_content,
               'msg_content': msg_content}
        return [msg]
    elif msg_type == 'content_ext':
        i_msg_type, msg_content, num_msg_content = generate_content_ext_messages(request_id, transform_id, workload_id, work, files=files,
                                                                                 relation_type=relation_type,
                                                                                 input_output_maps=input_output_maps)
        msg = {'msg_type': i_msg_type,
               'status': MessageStatus.New,
               'source': MessageSource.Carrier,
               'destination': MessageDestination.ContentExt,
               'request_id': request_id,
               'workload_id': workload_id,
               'transform_id': transform_id,
               'num_contents': num_msg_content,
               'msg_content': msg_content}
        return [msg]
    elif msg_type == 'collection':
        msg_type_contents = []
        for coll in files:
            msg_type_content = generate_collection_messages(request_id, transform_id, workload_id, work, coll, relation_type=relation_type)
            msg_type_contents.append(msg_type_content)

        msgs = []
        for i_msg_type, msg_content, num_msg_content in msg_type_contents:
            msg = {'msg_type': i_msg_type,
                   'status': MessageStatus.New,
                   'source': MessageSource.Carrier,
                   'destination': MessageDestination.Outside,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'transform_id': transform_id,
                   'num_contents': num_msg_content,
                   'msg_content': msg_content}
            msgs.append(msg)
        return msgs
    elif msg_type == 'work':
        # link collections
        input_collections = work.get_input_collections()
        output_collections = work.get_output_collections()
        log_collections = work.get_log_collections()

        msg_type_contents = []
        msg_type_content = generate_work_messages(request_id, transform_id, workload_id, work, relation_type='input')
        msg_type_contents.append(msg_type_content)
        for coll in input_collections:
            msg_type_content = generate_collection_messages(request_id, transform_id, workload_id, work, coll, relation_type='input')
            msg_type_contents.append(msg_type_content)
        for coll in output_collections:
            msg_type_content = generate_collection_messages(request_id, transform_id, workload_id, work, coll, relation_type='output')
            msg_type_contents.append(msg_type_content)
        for coll in log_collections:
            msg_type_content = generate_collection_messages(request_id, transform_id, workload_id, work, coll, relation_type='log')
            msg_type_contents.append(msg_type_content)

        msgs = []
        for i_msg_type, msg_content, num_msg_content in msg_type_contents:
            msg = {'msg_type': i_msg_type,
                   'status': MessageStatus.New,
                   'source': MessageSource.Carrier,
                   'destination': MessageDestination.Outside,
                   'request_id': request_id,
                   'workload_id': workload_id,
                   'transform_id': transform_id,
                   'num_contents': num_msg_content,
                   'msg_content': msg_content}
            msgs.append(msg)
        return msgs


def handle_new_processing(processing, agent_attributes, logger=None, log_prefix=''):
    logger = get_logger(logger)

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)
    transform_id = processing['transform_id']

    status, workload_id, errors = work.submit_processing(processing)
    logger.info(log_prefix + "submit_processing (status: %s, workload_id: %s, errors: %s)" % (status, workload_id, errors))

    if not status:
        logger.error(log_prefix + "Failed to submit processing (status: %s, workload_id: %s, errors: %s)" % (status, workload_id, errors))
        return False, processing, [], [], [], [], errors

    ret_msgs = []
    new_contents = []
    new_input_dependency_contents = []
    update_collections = []
    if proc.workload_id:
        processing['workload_id'] = proc.workload_id
        input_collections = work.get_input_collections()
        output_collections = work.get_output_collections()
        log_collections = work.get_log_collections()
        for coll in input_collections + output_collections + log_collections:
            u_coll = {'coll_id': coll.coll_id, 'workload_id': proc.workload_id}
            update_collections.append(u_coll)

    if proc.submitted_at:
        input_output_maps = get_input_output_maps(transform_id, work)
        new_input_output_maps = work.get_new_input_output_maps(input_output_maps)
        request_id = processing['request_id']
        transform_id = processing['transform_id']
        workload_id = processing['workload_id']
        ret_new_contents = get_new_contents(request_id, transform_id, workload_id, new_input_output_maps, logger=logger, log_prefix=log_prefix)
        new_input_contents, new_output_contents, new_log_contents, new_input_dependency_contents = ret_new_contents
        # new_contents = new_input_contents + new_output_contents + new_log_contents + new_input_dependency_contents
        new_contents = new_input_contents + new_output_contents + new_log_contents

        # not generate new messages
        # if new_input_contents:
        #     msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file', files=new_input_contents, relation_type='input')
        #     ret_msgs = ret_msgs + msgs
        # if new_output_contents:
        #     msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file', files=new_input_contents, relation_type='output')
        #     ret_msgs = ret_msgs + msgs
    return True, processing, update_collections, new_contents, new_input_dependency_contents, ret_msgs, errors


def get_updated_contents_by_request(request_id, transform_id, workload_id, work, terminated=False, input_output_maps=None,
                                    logger=None, log_prefix=''):
    logger = get_logger(logger)

    status_to_check = [ContentStatus.Available, ContentStatus.FakeAvailable, ContentStatus.FinalFailed,
                       ContentStatus.Missing, ContentStatus.Failed, ContentStatus.Lost,
                       ContentStatus.Deleted]
    contents = core_catalog.get_contents_by_request_transform(request_id=request_id, transform_id=transform_id,
                                                              status=status_to_check, status_updated=True)
    updated_contents, updated_contents_full_input, updated_contents_full_output = [], [], []
    updated_contents_full_input_deps = []
    for content in contents:
        if (content['status'] != content['substatus']) and content['substatus'] in status_to_check:
            u_content = {'content_id': content['content_id'],
                         'status': content['substatus']}
            updated_contents.append(u_content)
            if content['content_relation_type'] == ContentRelationType.Output:
                updated_contents_full_output.append(content)
            elif content['content_relation_type'] == ContentRelationType.Input:
                updated_contents_full_input.append(content)
            elif content['content_relation_type'] == ContentRelationType.InputDependency:
                updated_contents_full_input_deps.append(content)
    # logger.debug(log_prefix + "get_updated_contents_by_request: updated_contents[:3]: %s" % str(updated_contents[:3]))
    return updated_contents, updated_contents_full_input, updated_contents_full_output, updated_contents_full_input_deps


def get_updated_contents_by_input_output_maps(input_output_maps=None, logger=None, log_prefix=''):
    updated_contents, updated_contents_full_input, updated_contents_full_output = [], [], []
    updated_contents_full_input_deps = []
    new_update_contents = []

    status_to_check = [ContentStatus.Available, ContentStatus.FakeAvailable, ContentStatus.FinalFailed,
                       ContentStatus.Missing, ContentStatus.Failed, ContentStatus.Lost,
                       ContentStatus.Deleted]

    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        # logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        for content in inputs:
            if (content['status'] != content['substatus']) and content['substatus'] in status_to_check:
                u_content = {'content_id': content['content_id'],
                             'status': content['substatus']}
                updated_contents.append(u_content)
                u_content_substatus = {'content_id': content['content_id'],
                                       'substatus': content['substatus'],
                                       'request_id': content['request_id'],
                                       'transform_id': content['transform_id'],
                                       'workload_id': content['workload_id'],
                                       'coll_id': content['coll_id']}
                new_update_contents.append(u_content_substatus)
                updated_contents_full_input.append(content)
        for content in outputs:
            if (content['status'] != content['substatus']) and content['substatus'] in status_to_check:
                u_content = {'content_id': content['content_id'],
                             'status': content['substatus']}
                updated_contents.append(u_content)
                u_content_substatus = {'content_id': content['content_id'],
                                       'substatus': content['substatus'],
                                       'request_id': content['request_id'],
                                       'transform_id': content['transform_id'],
                                       'workload_id': content['workload_id'],
                                       'coll_id': content['coll_id']}
                new_update_contents.append(u_content_substatus)
                updated_contents_full_output.append(content)
        for content in inputs_dependency:
            if (content['status'] != content['substatus']) and content['substatus'] in status_to_check:
                u_content = {'content_id': content['content_id'],
                             'status': content['substatus']}
                updated_contents.append(u_content)
                updated_contents_full_input_deps.append(content)

        input_content_update_status = None
        if is_all_contents_available(inputs_dependency):
            input_content_update_status = ContentStatus.Available
        elif is_all_contents_terminated(inputs_dependency):
            input_content_update_status = ContentStatus.Missing
        if input_content_update_status:
            for content in inputs:
                if content['substatus'] != input_content_update_status:
                    u_content = {'content_id': content['content_id'],
                                 'status': input_content_update_status,
                                 'substatus': input_content_update_status}
                    updated_contents.append(u_content)
                    content['status'] = input_content_update_status
                    content['substatus'] = input_content_update_status
                    updated_contents_full_input.append(content)
                    u_content_substatus = {'content_id': content['content_id'],
                                           'substatus': content['substatus'],
                                           'request_id': content['request_id'],
                                           'transform_id': content['transform_id'],
                                           'workload_id': content['workload_id'],
                                           'coll_id': content['coll_id']}
                    new_update_contents.append(u_content_substatus)

        output_content_update_status = None
        if is_all_contents_available(inputs):
            # wait for the job to finish
            pass
        elif is_all_contents_terminated_but_not_available(inputs):
            output_content_update_status = ContentStatus.Missing
        if output_content_update_status:
            for content in outputs:
                if content['substatus'] != output_content_update_status:
                    u_content = {'content_id': content['content_id'],
                                 'status': output_content_update_status,
                                 'substatus': output_content_update_status}
                    updated_contents.append(u_content)
                    content['status'] = output_content_update_status
                    content['substatus'] = output_content_update_status
                    updated_contents_full_output.append(content)
                    u_content_substatus = {'content_id': content['content_id'],
                                           'substatus': content['substatus'],
                                           'request_id': content['request_id'],
                                           'transform_id': content['transform_id'],
                                           'workload_id': content['workload_id'],
                                           'coll_id': content['coll_id']}
                    new_update_contents.append(u_content_substatus)
    return updated_contents, updated_contents_full_input, updated_contents_full_output, updated_contents_full_input_deps, new_update_contents


def get_transform_dependency_map(transform_id, logger=None, log_prefix=''):
    cache = get_redis_cache()
    transform_dependcy_map_key = "transform_dependcy_map_%s" % transform_id
    transform_dependcy_map = cache.get(transform_dependcy_map_key, default={})
    return transform_dependcy_map


def set_transform_dependency_map(transform_id, transform_dependcy_map, logger=None, log_prefix=''):
    cache = get_redis_cache()
    transform_dependcy_map_key = "transform_dependcy_map_%s" % transform_id
    cache.set(transform_dependcy_map_key, transform_dependcy_map)


def get_content_dependcy_map(request_id, logger=None, log_prefix=''):
    cache = get_redis_cache()
    content_dependcy_map_key = "request_content_dependcy_map_%s" % request_id
    content_dependcy_map = cache.get(content_dependcy_map_key, default={})

    request_dependcy_map_key = "request_dependcy_map_%s" % request_id
    request_dependcy_map = cache.get(request_dependcy_map_key, default=[])

    collection_dependcy_map_key = "request_collections_dependcy_map_%s" % request_id
    collection_dependcy_map = cache.get(collection_dependcy_map_key, default=[])

    return content_dependcy_map, request_dependcy_map, collection_dependcy_map


def set_content_dependcy_map(request_id, content_dependcy_map, request_dependcy_map,
                             collection_dependcy_map, logger=None, log_prefix=''):
    cache = get_redis_cache()
    content_dependcy_map_key = "request_content_dependcy_map_%s" % request_id
    cache.set(content_dependcy_map_key, content_dependcy_map)

    request_dependcy_map_key = "request_dependcy_map_%s" % request_id
    cache.set(request_dependcy_map_key, request_dependcy_map)

    collection_dependcy_map_key = "request_collections_dependcy_map_%s" % request_id
    cache.set(collection_dependcy_map_key, collection_dependcy_map)


def get_content_status_map(request_id, logger=None, log_prefix=''):
    cache = get_redis_cache()
    content_status_map_key = "request_content_status_map_%s" % request_id
    content_status_map = cache.get(content_status_map_key, default={})
    return content_status_map


def set_content_status_map(request_id, content_status_map, logger=None, log_prefix=''):
    cache = get_redis_cache()
    content_status_map_key = "request_content_status_map_%s" % request_id
    cache.set(content_status_map_key, content_status_map)


def get_input_dependency_map_by_request(request_id, transform_id, workload_id, work, logger=None, log_prefix=''):
    logger = get_logger(logger)

    content_dependcy_map, request_dependcy_map, collection_dependcy_map = get_content_dependcy_map(request_id, logger=logger, log_prefix=log_prefix)
    content_status_map = get_content_status_map(request_id, logger=logger, log_prefix=log_prefix)

    transform_dependcy_maps = {}

    refresh = False
    if not content_dependcy_map or not content_status_map:
        refresh = True
    elif transform_id and transform_id not in request_dependcy_map:
        refresh = True
    elif work:
        output_collections = work.get_output_collections()
        for coll in output_collections:
            if coll.coll_id not in collection_dependcy_map:
                refresh = True

    for tf_id in request_dependcy_map:
        transform_dependcy_maps[str(tf_id)] = get_transform_dependency_map(tf_id, logger=logger, log_prefix=log_prefix)
        if not transform_dependcy_maps[str(tf_id)]:
            refresh = True

    if refresh:
        logger.debug(log_prefix + "refresh content_dependcy_map")
        content_dependcy_map = {}
        request_dependcy_map = []
        collection_dependcy_map = []
        content_status_map = {}
        transform_dependcy_maps = {}

        content_output_name2id = {}
        content_input_deps = []

        contents = core_catalog.get_contents_by_request_transform(request_id=request_id)
        # logger.debug("contents: ", contents)
        for content in contents:
            if content['transform_id'] not in request_dependcy_map:
                request_dependcy_map.append(content['transform_id'])
            if content['coll_id'] not in collection_dependcy_map:
                collection_dependcy_map.append(content['coll_id'])

            content_status_map[str(content['content_id'])] = content['substatus'].value

            str_tf_id = str(content['transform_id'])
            str_map_id = str(content['map_id'])
            if str_tf_id not in transform_dependcy_maps:
                transform_dependcy_maps[str_tf_id] = get_transform_dependency_map(str_tf_id, logger=logger, log_prefix=log_prefix)
            if str_map_id not in transform_dependcy_maps[str_tf_id]:
                transform_dependcy_maps[str_tf_id][str_map_id] = {'inputs': [], 'outputs': [], 'input_deps': []}

            if content['content_relation_type'] == ContentRelationType.Output:
                if content['coll_id'] not in content_output_name2id:
                    content_output_name2id[content['coll_id']] = {}
                    collection_dependcy_map.append(content['coll_id'])
                content_output_name2id[content['coll_id']][content['name']] = content
                # content_id, status
                transform_dependcy_maps[str_tf_id][str_map_id]['outputs'].append(content['content_id'])
            elif content['content_relation_type'] == ContentRelationType.InputDependency:
                content_input_deps.append(content)
                # content_id, status
                transform_dependcy_maps[str_tf_id][str_map_id]['input_deps'].append(content['content_id'])
            elif content['content_relation_type'] == ContentRelationType.Input:
                # content_id, status
                transform_dependcy_maps[str_tf_id][str_map_id]['inputs'].append(content['content_id'])
        # logger.debug("content_output_name2id: ", content_output_name2id)

        for content in content_input_deps:
            dep_coll_id = content['coll_id']
            if dep_coll_id not in content_output_name2id:
                logger.warn(log_prefix + "dep_coll_id: %s contents are not added yet" % dep_coll_id)
            else:
                dep_content = content_output_name2id[dep_coll_id].get(content['name'], None)
                if dep_content:
                    dep_content_id = str(dep_content['content_id'])
                    if dep_content_id not in content_dependcy_map:
                        content_dependcy_map[dep_content_id] = []
                    content_dependcy_map[dep_content_id].append((content['content_id'], content['transform_id'], content['map_id']))
                else:
                    logger.error(log_prefix + "Failed to find input dependcy for content_id: %s" % content['content_id'])

        set_content_dependcy_map(request_id, content_dependcy_map, request_dependcy_map,
                                 collection_dependcy_map, logger=logger, log_prefix=log_prefix)
        for str_tf_id in transform_dependcy_maps:
            set_transform_dependency_map(str_tf_id, transform_dependcy_maps[str_tf_id], logger=logger, log_prefix=log_prefix)
        set_content_status_map(request_id, content_status_map, logger=logger, log_prefix=log_prefix)

    return content_dependcy_map, transform_dependcy_maps, content_status_map


def get_content_status_with_status_map(content_ids, content_status_map):
    content_id_status = []
    for content_id in content_ids:
        status_value = content_status_map[str(content_id)]
        status = ContentStatus(status_value)
        content_id_status.append((content_id, status))
    return content_id_status


def trigger_release_inputs_no_deps(request_id, transform_id, workload_id, work, input_output_maps, logger=None, log_prefix=''):
    logger = get_logger(logger)

    update_contents = []
    update_input_contents_full = {}
    update_input_contents_full[transform_id] = []

    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        # outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        # logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        if not inputs_dependency:
            for content in inputs:
                if content['substatus'] != ContentStatus.Available:
                    u_content = {'content_id': content['content_id'],
                                 # 'status': ContentStatus.Available,
                                 'substatus': ContentStatus.Available}
                    update_contents.append(u_content)
                    content['status'] = ContentStatus.Available
                    content['substatus'] = ContentStatus.Available
                    update_input_contents_full[transform_id].append(content)
    return update_contents, update_input_contents_full


def trigger_release_inputs(request_id, transform_id, workload_id, work, updated_contents_full_output, updated_contents_full_input,
                           updated_contents_full_input_deps, input_output_maps, logger=None, log_prefix=''):
    logger = get_logger(logger)

    status_to_check = [ContentStatus.Available, ContentStatus.FakeAvailable, ContentStatus.FinalFailed, ContentStatus.Missing]
    # status_to_check_fake = [ContentStatus.FakeAvailable, ContentStatus.Missing]

    update_contents = []
    update_contents_status = {}
    update_contents_status_name = {}
    update_input_contents_full = {}
    update_input_contents_full[transform_id] = []

    for status in status_to_check:
        update_contents_status[status.name] = []
        update_contents_status_name[status.name] = status

    for content in updated_contents_full_output:
        # update the status
        # u_content = {'content_id': content['content_id'], 'status': content['substatus']}
        # update_contents.append(u_content)

        if content['substatus'] in status_to_check:
            update_contents_status[content['substatus'].name].append(content['content_id'])

    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        # logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        input_content_update_status = None
        if is_all_contents_available(inputs_dependency):
            input_content_update_status = ContentStatus.Available
        elif is_all_contents_terminated(inputs_dependency):
            input_content_update_status = ContentStatus.Missing
        if input_content_update_status:
            for content in inputs_dependency:
                # u_content = {'content_id': content['content_id'], 'status': content['substatus'])
                # update_contents.append(u_content)
                pass
            for content in inputs:
                u_content = {'content_id': content['content_id'],
                             'substatus': input_content_update_status}
                update_contents.append(u_content)
                content['status'] = input_content_update_status
                content['substatus'] = input_content_update_status
                update_input_contents_full[transform_id].append(content)

        output_content_update_status = None
        if is_all_contents_available(inputs):
            # wait for the job to finish
            # for content in inputs:
            #     u_content = {'content_id': content['content_id'], 'status': content['substatus'])
            #     update_contents.append(u_content)
            pass
        elif is_all_contents_terminated_but_not_available(inputs):
            # for content in inputs:
            #     u_content = {'content_id': content['content_id'], 'status': content['substatus'])
            #     update_contents.append(u_content)
            pass
            output_content_update_status = ContentStatus.Missing
        if output_content_update_status:
            for content in outputs:
                u_content = {'content_id': content['content_id'],
                             'substatus': output_content_update_status}
                update_contents.append(u_content)

    return update_contents, update_input_contents_full, update_contents_status_name, update_contents_status


def poll_missing_outputs(input_output_maps):
    content_updates_missing, updated_contents_full_missing = [], []

    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        # inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        # logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        content_update_status = None
        if is_all_contents_terminated_but_not_available(inputs):
            content_update_status = ContentStatus.Missing

            for content in outputs:
                content['substatus'] = content_update_status
                if content['status'] != content['substatus']:
                    u_content = {'content_id': content['content_id'],
                                 'substatus': content['substatus']}

                    content_updates_missing.append(u_content)
                    updated_contents_full_missing.append(content)

    return content_updates_missing, updated_contents_full_missing


def handle_update_processing(processing, agent_attributes, logger=None, log_prefix=''):
    logger = get_logger(logger)

    ret_msgs = []
    new_contents = []

    request_id = processing['request_id']
    transform_id = processing['transform_id']
    workload_id = processing['workload_id']

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)

    input_output_maps = get_input_output_maps(transform_id, work)
    logger.debug(log_prefix + "get_input_output_maps: len: %s" % len(input_output_maps))
    logger.debug(log_prefix + "get_input_output_maps.keys[:3]: %s" % str(list(input_output_maps.keys())[:3]))

    new_input_output_maps = work.get_new_input_output_maps(input_output_maps)
    logger.debug(log_prefix + "get_new_input_output_maps: len: %s" % len(new_input_output_maps))
    logger.debug(log_prefix + "get_new_input_output_maps.keys[:3]: %s" % str(list(new_input_output_maps.keys())[:3]))

    if work.require_ext_contents():
        contents_ext = get_ext_contents(transform_id, work)
        job_info_maps = core_catalog.get_contents_ext_maps()
        ret_poll_processing = work.poll_processing_updates(processing, input_output_maps, contents_ext=contents_ext,
                                                           job_info_maps=job_info_maps, log_prefix=log_prefix)
        process_status, content_updates, new_input_output_maps1, updated_contents_full, parameters, new_contents_ext, update_contents_ext = ret_poll_processing
    else:
        ret_poll_processing = work.poll_processing_updates(processing, input_output_maps, log_prefix=log_prefix)
        new_contents_ext, update_contents_ext = [], []
        process_status, content_updates, new_input_output_maps1, updated_contents_full, parameters = ret_poll_processing

    new_input_output_maps.update(new_input_output_maps1)
    logger.debug(log_prefix + "poll_processing_updates process_status: %s" % process_status)
    logger.debug(log_prefix + "poll_processing_updates content_updates[:3]: %s" % content_updates[:3])
    logger.debug(log_prefix + "poll_processing_updates new_input_output_maps1.keys[:3]: %s" % (list(new_input_output_maps1.keys())[:3]))
    logger.debug(log_prefix + "poll_processing_updates updated_contents_full[:3]: %s" % (updated_contents_full[:3]))

    ret_new_contents = get_new_contents(request_id, transform_id, workload_id, new_input_output_maps)
    new_input_contents, new_output_contents, new_log_contents, new_input_dependency_contents = ret_new_contents
    # new_contents = new_input_contents + new_output_contents + new_log_contents + new_input_dependency_contents
    new_contents = new_input_contents + new_output_contents + new_log_contents

    content_updates_missing, updated_contents_full_missing = poll_missing_outputs(input_output_maps)

    if new_input_contents:
        msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file',
                                 files=new_input_contents, relation_type='input')
        ret_msgs = ret_msgs + msgs
    if new_output_contents:
        msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file',
                                 files=new_output_contents, relation_type='output')
        ret_msgs = ret_msgs + msgs

    updated_contents_full = updated_contents_full + updated_contents_full_missing
    if updated_contents_full:
        msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file',
                                 files=updated_contents_full, relation_type='output')
        ret_msgs = ret_msgs + msgs

    return process_status, new_contents, new_input_dependency_contents, ret_msgs, content_updates + content_updates_missing, parameters, new_contents_ext, update_contents_ext


def handle_trigger_processing(processing, agent_attributes, trigger_new_updates=False, logger=None, log_prefix=''):
    logger = get_logger(logger)

    ret_msgs = []
    content_updates = []
    ret_update_transforms = []

    request_id = processing['request_id']
    transform_id = processing['transform_id']
    workload_id = processing['workload_id']

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)

    if (not work.use_dependency_to_release_jobs()) or workload_id is None:
        return processing['substatus'], [], [], {}, {}, {}, []
    else:
        if trigger_new_updates:
            # delete information in the contents_update table, to invoke the trigger.
            # ret_update_transforms = core_catalog.delete_contents_update(request_id=request_id, transform_id=transform_id)
            # logger.debug(log_prefix + "delete_contents_update: %s" % str(ret_update_transforms))
            pass

        logger.debug(log_prefix + "update_contents_from_others_by_dep_id")
        core_catalog.update_contents_from_others_by_dep_id(request_id=request_id, transform_id=transform_id)
        logger.debug(log_prefix + "update_contents_from_others_by_dep_id done")

        input_output_maps = get_input_output_maps(transform_id, work)
        logger.debug(log_prefix + "input_output_maps.keys[:2]: %s" % str(list(input_output_maps.keys())[:2]))

        updated_contents_ret = get_updated_contents_by_input_output_maps(input_output_maps=input_output_maps, logger=logger, log_prefix=log_prefix)

        updated_contents, updated_contents_full_input, updated_contents_full_output, updated_contents_full_input_deps, new_update_contents = updated_contents_ret
        logger.debug(log_prefix + "handle_trigger_processing: updated_contents[:3] %s" % (updated_contents[:3]))

        if updated_contents_full_input:
            # if the content is updated by receiver, here is the place to broadcast the messages
            msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file',
                                     files=updated_contents_full_input, relation_type='input')
            ret_msgs = ret_msgs + msgs
        if updated_contents_full_output:
            # if the content is updated by receiver, here is the place to broadcast the messages
            msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='file',
                                     files=updated_contents_full_output, relation_type='output')
            ret_msgs = ret_msgs + msgs

        content_updates = content_updates + updated_contents

    # update_dep_contents_status_name = {}
    # update_dep_contents_status = {}
    # for content in new_update_contents:
    #     if content['substatus'] not in update_dep_contents_status_name:
    #         update_dep_contents_status_name[content['substatus'].name] = content['substatus']
    #         update_dep_contents_status[content['substatus'].name] = []
    #     update_dep_contents_status[content['substatus'].name].append(content['content_id'])

    # return processing['substatus'], content_updates, ret_msgs, {}, {}, {}, new_update_contents, ret_update_transforms
    # return processing['substatus'], content_updates, ret_msgs, {}, update_dep_contents_status_name, update_dep_contents_status, [], ret_update_transforms
    # return processing['substatus'], content_updates, ret_msgs, {}, {}, {}, [], ret_update_transforms
    return processing['substatus'], content_updates, ret_msgs, {}, {}, {}, new_update_contents, ret_update_transforms


def get_content_status_from_panda_msg_status(status):
    status_map = {'starting': ContentStatus.New,
                  'running': ContentStatus.Processing,
                  'finished': ContentStatus.Available,
                  'failed': ContentStatus.Failed}
    if status in status_map:
        return status_map[status]
    return ContentStatus.New


def get_collection_id_transform_id_map(coll_id, request_id, request_ids=[]):
    cache = get_redis_cache()
    coll_tf_id_map_key = "collection_id_transform_id_map"
    coll_tf_id_map = cache.get(coll_tf_id_map_key, default={})

    if coll_id is None or coll_id not in coll_tf_id_map:
        if not request_ids:
            request_ids = []
        if request_id not in request_ids:
            request_ids.append(request_id)
        colls = core_catalog.get_collections_by_request_ids(request_ids)
        for coll in colls:
            coll_tf_id_map[coll['coll_id']] = (coll['request_id'], coll['transform_id'], coll['workload_id'])

        cache.set(coll_tf_id_map_key, coll_tf_id_map)

    if coll_id is None or coll_id not in coll_tf_id_map:
        return None, None, None
    return coll_tf_id_map[coll_id]


workload_id_lock = threading.Lock()


def get_workload_id_transform_id_map(workload_id, logger=None, log_prefix=''):
    cache = get_redis_cache()
    workload_id_transform_id_map_key = "all_worloadid2transformid_map"
    workload_id_transform_id_map = cache.get(workload_id_transform_id_map_key, default={})

    workload_id_transform_id_map_notexist_key = "all_worloadid2transformid_map_notexist"
    workload_id_transform_id_map_notexist = cache.get(workload_id_transform_id_map_notexist_key, default={})

    if type(workload_id_transform_id_map_notexist) in (list, tuple):
        workload_id_transform_id_map_notexist = {}

    workload_id_str = str(workload_id)
    if workload_id_str in workload_id_transform_id_map:
        return workload_id_transform_id_map[workload_id_str]

    if workload_id_str in workload_id_transform_id_map_notexist and workload_id_transform_id_map_notexist[workload_id_str] + 600 < time.time():
        return None

    # lock area
    workload_id_lock.acquire()

    workload_id_transform_id_map = cache.get(workload_id_transform_id_map_key, default={})
    workload_id_transform_id_map_notexist = cache.get(workload_id_transform_id_map_notexist_key, default={})

    if type(workload_id_transform_id_map_notexist) in (list, tuple):
        workload_id_transform_id_map_notexist = {}

    request_ids = []
    if not workload_id_transform_id_map or workload_id_str not in workload_id_transform_id_map or len(workload_id_transform_id_map[workload_id_str]) < 5:
        processing_status = [ProcessingStatus.New,
                             ProcessingStatus.Submitting, ProcessingStatus.Submitted,
                             ProcessingStatus.Running, ProcessingStatus.FinishedOnExec,
                             ProcessingStatus.Cancel, ProcessingStatus.FinishedOnStep,
                             ProcessingStatus.ToCancel, ProcessingStatus.Cancelling,
                             ProcessingStatus.ToSuspend, ProcessingStatus.Suspending,
                             ProcessingStatus.ToResume, ProcessingStatus.Resuming,
                             ProcessingStatus.ToExpire, ProcessingStatus.Expiring,
                             ProcessingStatus.ToFinish, ProcessingStatus.ToForceFinish]

        procs = core_processings.get_processings_by_status(status=processing_status)
        for proc in procs:
            processing = proc['processing_metadata']['processing']
            work = processing.work
            if work.use_dependency_to_release_jobs():
                workload_id_transform_id_map[str(proc['workload_id'])] = (proc['request_id'],
                                                                          proc['transform_id'],
                                                                          proc['processing_id'],
                                                                          proc['status'].value,
                                                                          proc['substatus'].value)
                if proc['request_id'] not in request_ids:
                    request_ids.append(proc['request_id'])

        cache.set(workload_id_transform_id_map_key, workload_id_transform_id_map)

        for key in workload_id_transform_id_map:
            if key in workload_id_transform_id_map_notexist:
                del workload_id_transform_id_map_notexist[key]

    # renew the collection to transform map
    if request_ids:
        get_collection_id_transform_id_map(coll_id=None, request_id=request_ids[0], request_ids=request_ids)

    keys = list(workload_id_transform_id_map_notexist.keys())
    for key in keys:
        if workload_id_transform_id_map_notexist[key] + 7200 < time.time():
            del workload_id_transform_id_map_notexist[key]

    cache.set(workload_id_transform_id_map_notexist_key, workload_id_transform_id_map_notexist)
    # for tasks running in some other instances
    if workload_id_str not in workload_id_transform_id_map:
        if workload_id_str not in workload_id_transform_id_map_notexist:
            workload_id_transform_id_map_notexist[workload_id_str] = time.time()
            cache.set(workload_id_transform_id_map_notexist_key, workload_id_transform_id_map_notexist)

        workload_id_lock.release()
        return None
    else:
        workload_id_lock.release()

    return workload_id_transform_id_map[workload_id_str]


def get_input_name_content_id_map(request_id, workload_id, transform_id):
    cache = get_redis_cache()
    input_name_content_id_map_key = "transform_input_contentid_map_%s" % transform_id
    input_name_content_id_map = cache.get(input_name_content_id_map_key, default={})

    if not input_name_content_id_map:
        contents = core_catalog.get_contents_by_request_transform(request_id=request_id, transform_id=transform_id)
        input_name_content_id_map = {}
        for content in contents:
            if content['content_relation_type'] == ContentRelationType.Output:
                input_name_content_id_map[content['name']] = content['content_id']

        cache.set(input_name_content_id_map_key, input_name_content_id_map)
    return input_name_content_id_map


def get_jobid_content_id_map(request_id, workload_id, transform_id, job_id, inputs):
    cache = get_redis_cache()
    jobid_content_id_map_key = "transform_jobid_contentid_map_%s" % transform_id
    jobid_content_id_map = cache.get(jobid_content_id_map_key, default={})

    to_update_jobid = False
    job_id = str(job_id)
    if not jobid_content_id_map or job_id not in jobid_content_id_map:
        to_update_jobid = True
        input_name_content_id_map = get_input_name_content_id_map(request_id, workload_id, transform_id)
        for ip in inputs:
            if ':' in ip:
                pos = ip.find(":")
                ip = ip[pos + 1:]
            if ip in input_name_content_id_map:
                content_id = input_name_content_id_map[ip]
                jobid_content_id_map[job_id] = content_id
                break

        cache.set(jobid_content_id_map_key, jobid_content_id_map)
    return jobid_content_id_map, to_update_jobid


def get_content_id_from_job_id(request_id, workload_id, transform_id, job_id, inputs):
    jobid_content_id_map, to_update_jobid = get_jobid_content_id_map(request_id, workload_id, transform_id, job_id, inputs)

    if str(job_id) in jobid_content_id_map:
        content_id = jobid_content_id_map[str(job_id)]
    else:
        content_id = None
    return content_id, to_update_jobid


pending_lock = threading.Lock()


def whether_to_process_pending_workload_id(workload_id, logger=None, log_prefix=''):
    cache = get_redis_cache()
    processed_pending_workload_id_map_key = "processed_pending_workload_id_map"
    processed_pending_workload_id_map = cache.get(processed_pending_workload_id_map_key, default={})
    processed_pending_workload_id_map_time_key = "processed_pending_workload_id_map_time"
    processed_pending_workload_id_map_time = cache.get(processed_pending_workload_id_map_time_key, default=None)

    workload_id = str(workload_id)
    if workload_id in processed_pending_workload_id_map:
        return False

    # lock area
    pending_lock.acquire()
    processed_pending_workload_id_map = cache.get(processed_pending_workload_id_map_key, default={})

    processed_pending_workload_id_map[workload_id] = time.time()
    if processed_pending_workload_id_map_time is None or processed_pending_workload_id_map_time + 86400 < time.time():
        cache.set(processed_pending_workload_id_map_time_key, int(time.time()), expire_seconds=86400)

        keys = list(processed_pending_workload_id_map.keys())
        for workload_id in keys:
            if processed_pending_workload_id_map[workload_id] + 86400 < time.time():
                del processed_pending_workload_id_map[workload_id]

    cache.set(processed_pending_workload_id_map_key, processed_pending_workload_id_map, expire_seconds=86400)
    pending_lock.release()
    return True


update_processing_lock = threading.Lock()


def whether_to_update_processing(processing_id, interval=300):
    cache = get_redis_cache()
    ret = False

    update_processing_lock.acquire()
    update_processing_map_key = "update_processing_map"
    update_processing_map = cache.get(update_processing_map_key, default={})

    processing_id_str = str(processing_id)
    if processing_id_str not in update_processing_map or update_processing_map[processing_id_str] + interval < time.time():
        update_processing_map[processing_id_str] = time.time()
        ret = True

    keys = list(update_processing_map.keys())
    for key in keys:
        if update_processing_map[key] + 86400 < time.time():
            del update_processing_map[key]

    cache.set(update_processing_map_key, default=update_processing_map)
    update_processing_lock.release()
    return ret


def handle_messages_processing(messages, logger=None, log_prefix='', update_processing_interval=300):
    logger = get_logger(logger)
    if not log_prefix:
        log_prefix = "<Message>"

    update_processings = []
    terminated_processings = []
    update_contents = []

    for ori_msg in messages:
        msg = json.loads(ori_msg)
        if 'taskid' not in msg or not msg['taskid']:
            continue

        logger.debug(log_prefix + "Received message: %s" % str(ori_msg))

        if msg['msg_type'] in ['task_status']:
            workload_id = msg['taskid']
            status = msg['status']
            if status in ['pending']:   # 'prepared'
                ret_req_tf_pr_id = get_workload_id_transform_id_map(workload_id, logger=logger, log_prefix=log_prefix)
                if not ret_req_tf_pr_id:
                    # request is submitted by some other instances
                    logger.debug(log_prefix + "No matched workload_id, discard message: %s" % str(ori_msg))
                    continue

                logger.debug(log_prefix + "(request_id, transform_id, processing_id, status, substatus): %s" % str(ret_req_tf_pr_id))
                req_id, tf_id, processing_id, r_status, r_substatus = ret_req_tf_pr_id
                if whether_to_process_pending_workload_id(workload_id, logger=logger, log_prefix=log_prefix):
                    # new_processings.append((req_id, tf_id, processing_id, workload_id, status))
                    if processing_id not in update_processings:
                        update_processings.append(processing_id)
                        logger.debug(log_prefix + "Add to update processing: %s" % str(processing_id))
                else:
                    logger.debug(log_prefix + "Processing %s is already processed, not add it to update processing" % (str(processing_id)))
            elif status in ['finished', 'done']:
                ret_req_tf_pr_id = get_workload_id_transform_id_map(workload_id, logger=logger, log_prefix=log_prefix)
                if not ret_req_tf_pr_id:
                    # request is submitted by some other instances
                    logger.debug(log_prefix + "No matched workload_id, discard message: %s" % str(ori_msg))
                    continue

                logger.debug(log_prefix + "(request_id, transform_id, processing_id, status, substatus): %s" % str(ret_req_tf_pr_id))
                req_id, tf_id, processing_id, r_status, r_substatus = ret_req_tf_pr_id
                # update_processings.append((processing_id, status))
                if processing_id not in update_processings:
                    terminated_processings.append(processing_id)
                    logger.debug(log_prefix + "Add to terminated processing: %s" % str(processing_id))

        if msg['msg_type'] in ['job_status']:
            workload_id = msg['taskid']
            job_id = msg['jobid']
            status = msg['status']
            inputs = msg['inputs']
            if inputs and status in ['finished']:
                ret_req_tf_pr_id = get_workload_id_transform_id_map(workload_id, logger=logger, log_prefix=log_prefix)
                if not ret_req_tf_pr_id:
                    # request is submitted by some other instances
                    logger.debug(log_prefix + "No matched workload_id, discard message: %s" % str(ori_msg))
                    continue

                logger.debug(log_prefix + "(request_id, transform_id, processing_id, status, substatus): %s" % str(ret_req_tf_pr_id))

                req_id, tf_id, processing_id, r_status, r_substatus = ret_req_tf_pr_id
                content_id, to_update_jobid = get_content_id_from_job_id(req_id, workload_id, tf_id, job_id, inputs)
                if content_id:
                    if to_update_jobid:
                        u_content = {'content_id': content_id,
                                     # 'status': get_content_status_from_panda_msg_status(status),
                                     'substatus': get_content_status_from_panda_msg_status(status),
                                     'content_metadata': {'panda_id': job_id}}
                    else:
                        u_content = {'content_id': content_id,
                                     'substatus': get_content_status_from_panda_msg_status(status)}
                        #             # 'status': get_content_status_from_panda_msg_status(status)}

                    update_contents.append(u_content)
                    # if processing_id not in update_processings:
                    if processing_id not in update_processings and whether_to_update_processing(processing_id, update_processing_interval):
                        update_processings.append(processing_id)
                        logger.debug(log_prefix + "Add to update processing: %s" % str(processing_id))

    return update_processings, terminated_processings, update_contents, []


def sync_collection_status(request_id, transform_id, workload_id, work, input_output_maps=None,
                           close_collection=False, force_close_collection=False, terminate=False):
    if input_output_maps is None:
        input_output_maps = get_input_output_maps(transform_id, work)

    all_updates_flushed = True
    coll_status = {}
    messages = []
    for map_id in input_output_maps:
        inputs = input_output_maps[map_id]['inputs'] if 'inputs' in input_output_maps[map_id] else []
        # inputs_dependency = input_output_maps[map_id]['inputs_dependency'] if 'inputs_dependency' in input_output_maps[map_id] else []
        outputs = input_output_maps[map_id]['outputs'] if 'outputs' in input_output_maps[map_id] else []
        logs = input_output_maps[map_id]['logs'] if 'logs' in input_output_maps[map_id] else []

        for content in inputs + outputs + logs:
            if content['coll_id'] not in coll_status:
                coll_status[content['coll_id']] = {'total_files': 0, 'processed_files': 0, 'processing_files': 0, 'bytes': 0,
                                                   'new_files': 0, 'failed_files': 0, 'missing_files': 0,
                                                   'ext_files': 0, 'processed_ext_files': 0, 'failed_ext_files': 0,
                                                   'missing_ext_files': 0}
            coll_status[content['coll_id']]['total_files'] += 1

            if content['status'] in [ContentStatus.Available, ContentStatus.Mapped,
                                     ContentStatus.Available.value, ContentStatus.Mapped.value,
                                     ContentStatus.FakeAvailable, ContentStatus.FakeAvailable.value]:
                coll_status[content['coll_id']]['processed_files'] += 1
                coll_status[content['coll_id']]['bytes'] += content['bytes']
            elif content['status'] in [ContentStatus.New]:
                coll_status[content['coll_id']]['new_files'] += 1
            elif content['status'] in [ContentStatus.Failed, ContentStatus.FinalFailed]:
                coll_status[content['coll_id']]['failed_files'] += 1
            elif content['status'] in [ContentStatus.Lost, ContentStatus.Deleted, ContentStatus.Missing]:
                coll_status[content['coll_id']]['missing_files'] += 1
            else:
                coll_status[content['coll_id']]['processing_files'] += 1

            if content['status'] != content['substatus']:
                all_updates_flushed = False

    all_ext_updated = True
    if work.require_ext_contents():
        all_ext_updated = False
        contents_ext = core_catalog.get_contents_ext(request_id=request_id, transform_id=transform_id)
        for content in contents_ext:
            coll_status[content['coll_id']]['ext_files'] += 1

            if content['status'] in [ContentStatus.Available, ContentStatus.Mapped,
                                     ContentStatus.Available.value, ContentStatus.Mapped.value,
                                     ContentStatus.FakeAvailable, ContentStatus.FakeAvailable.value]:
                coll_status[content['coll_id']]['processed_ext_files'] += 1
            elif content['status'] in [ContentStatus.Failed, ContentStatus.FinalFailed]:
                coll_status[content['coll_id']]['failed_ext_files'] += 1
            elif content['status'] in [ContentStatus.Lost, ContentStatus.Deleted, ContentStatus.Missing]:
                coll_status[content['coll_id']]['missing_ext_files'] += 1

    input_collections = work.get_input_collections(poll_externel=True)
    output_collections = work.get_output_collections()
    log_collections = work.get_log_collections()

    update_collections = []
    for coll in input_collections + output_collections + log_collections:
        if coll.coll_id in coll_status:
            if 'total_files' in coll.coll_metadata and coll.coll_metadata['total_files']:
                coll.total_files = coll.coll_metadata['total_files']
            else:
                coll.total_files = coll_status[coll.coll_id]['total_files']
            coll.processed_files = coll_status[coll.coll_id]['processed_files']
            coll.processing_files = coll_status[coll.coll_id]['processing_files']
            coll.bytes = coll_status[coll.coll_id]['bytes']
            coll.new_files = coll_status[coll.coll_id]['new_files']
            coll.failed_files = coll_status[coll.coll_id]['failed_files']
            coll.missing_files = coll_status[coll.coll_id]['missing_files']
            coll.ext_files = coll_status[coll.coll_id]['ext_files']
            coll.processed_ext_files = coll_status[coll.coll_id]['processed_ext_files']
            coll.failed_ext_files = coll_status[coll.coll_id]['failed_ext_files']
            coll.missing_ext_files = coll_status[coll.coll_id]['missing_ext_files']
        else:
            coll.total_files = 0
            coll.processed_files = 0
            coll.processing_files = 0
            coll.new_files = 0
            coll.failed_files = 0
            coll.missing_files = 0
            coll.ext_files = 0
            coll.processed_ext_files = 0
            coll.failed_ext_files = 0
            coll.missing_ext_files = 0

        u_coll = {'coll_id': coll.coll_id,
                  'total_files': coll.total_files,
                  'processed_files': coll.processed_files,
                  'processing_files': coll.processing_files,
                  'new_files': coll.new_files,
                  'failed_files': coll.failed_files,
                  'missing_files': coll.missing_files,
                  'bytes': coll.bytes,
                  'ext_files': coll.ext_files,
                  'processed_ext_files': coll.processed_ext_files,
                  'failed_ext_files': coll.failed_ext_files,
                  'missing_ext_files': coll.missing_ext_files}

        if coll in input_collections and (workload_id is not None):
            if coll.total_files == coll.processed_files + coll.failed_files + coll.missing_files:
                coll_db = core_catalog.get_collection(coll_id=coll.coll_id)
                coll.status = coll_db['status']
                if coll.status is not None and coll.status != CollectionStatus.Closed:
                    u_coll['status'] = CollectionStatus.Closed
                    u_coll['substatus'] = CollectionStatus.Closed
                    coll.status = CollectionStatus.Closed
                    coll.substatus = CollectionStatus.Closed

                    msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='collection', files=[coll], relation_type='input')
                    messages += msgs

        if terminate:
            if coll in output_collections and work.require_ext_contents():
                if coll.processed_files == coll.processed_ext_files and coll.failed_files == coll.failed_ext_files:
                    all_ext_updated = True
                if (force_close_collection or (close_collection and all_updates_flushed and all_ext_updated)
                   or coll.status == CollectionStatus.Closed):        # noqa W503
                    u_coll['status'] = CollectionStatus.Closed
                    u_coll['substatus'] = CollectionStatus.Closed
                    coll.status = CollectionStatus.Closed
                    coll.substatus = CollectionStatus.Closed
            elif force_close_collection or close_collection and all_updates_flushed or coll.status == CollectionStatus.Closed:
                u_coll['status'] = CollectionStatus.Closed
                u_coll['substatus'] = CollectionStatus.Closed
                coll.status = CollectionStatus.Closed
                coll.substatus = CollectionStatus.Closed

        update_collections.append(u_coll)
    return update_collections, all_updates_flushed, messages


def sync_work_status(request_id, transform_id, workload_id, work):
    input_collections = work.get_input_collections()
    output_collections = work.get_output_collections()
    log_collections = work.get_log_collections()

    is_all_collections_closed = True
    is_all_files_processed = True
    is_all_files_failed = True
    for coll in input_collections + output_collections + log_collections:
        if coll.status != CollectionStatus.Closed:
            is_all_collections_closed = False
    for coll in output_collections:
        if coll.total_files != coll.processed_files:
            is_all_files_processed = False
        if coll.processed_files > 0 or coll.total_files == coll.processed_files:
            is_all_files_failed = False

    if is_all_collections_closed:
        if is_all_files_processed:
            work.status = WorkStatus.Finished
        elif is_all_files_failed:
            work.status = WorkStatus.Failed
        else:
            work.status = WorkStatus.SubFinished


def sync_processing(processing, agent_attributes, terminate=False, logger=None, log_prefix=""):
    logger = get_logger()

    request_id = processing['request_id']
    transform_id = processing['transform_id']
    workload_id = processing['workload_id']

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)

    messages = []
    input_output_maps = get_input_output_maps(transform_id, work)
    update_collections, all_updates_flushed, msgs = sync_collection_status(request_id, transform_id, workload_id, work,
                                                                           input_output_maps=input_output_maps,
                                                                           close_collection=True, terminate=terminate)

    messages += msgs

    sync_work_status(request_id, transform_id, workload_id, work)
    logger.info(log_prefix + "sync_processing: work status: %s" % work.get_status())
    if terminate and work.is_terminated():
        msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='work')
        messages += msgs
        if work.is_finished():
            # processing['status'] = ProcessingStatus.Finished
            processing['status'] = processing['substatus']
        elif work.is_subfinished():
            processing['status'] = ProcessingStatus.SubFinished
        elif work.is_failed():
            processing['status'] = ProcessingStatus.Failed
        else:
            processing['status'] = ProcessingStatus.SubFinished

        if work.require_ext_contents():
            contents_ext = core_catalog.get_contents_ext(request_id=request_id, transform_id=transform_id)
            msgs = generate_messages(request_id, transform_id, workload_id, work, msg_type='content_ext', files=contents_ext,
                                     relation_type='output', input_output_maps=input_output_maps)
            messages += msgs

        if processing['status'] == ProcessingStatus.Terminating and is_process_terminated(processing['substatus']):
            processing['status'] = processing['substatus']

    return processing, update_collections, messages


def handle_abort_processing(processing, agent_attributes, logger=None, log_prefix=''):
    logger = get_logger(logger)

    # request_id = processing['request_id']
    # transform_id = processing['transform_id']
    # workload_id = processing['workload_id']

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)

    work.abort_processing(processing, log_prefix=log_prefix)

    # input_collections = work.get_input_collections()
    # output_collections = work.get_output_collections()
    # log_collections = work.get_log_collections()

    # input_output_maps = get_input_output_maps(transform_id, work)
    # update_collections, all_updates_flushed = sync_collection_status(request_id, transform_id, workload_id, work,
    #                                                                  input_output_maps=None, close_collection=True,
    #                                                                  force_close_collection=True)

    # for coll in input_collections + output_collections + log_collections:
    #     coll.status = CollectionStatus.Closed
    #     coll.substatus = CollectionStatus.Closed
    processing, update_collections, messages = sync_processing(processing, agent_attributes, terminate=True, logger=logger, log_prefix=log_prefix)
    update_contents = []

    # processing['status'] = ProcessingStatus.Cancelled
    return processing, update_collections, update_contents, messages


def reactive_contents(request_id, transform_id, workload_id, work, input_output_maps):
    updated_contents = []
    contents = core_catalog.get_contents_by_request_transform(request_id=request_id, transform_id=transform_id)
    for content in contents:
        if content['status'] not in [ContentStatus.Available, ContentStatus.Mapped,
                                     ContentStatus.Available.value, ContentStatus.Mapped.value,
                                     ContentStatus.FakeAvailable, ContentStatus.FakeAvailable.value]:
            u_content = {'content_id': content['content_id'],
                         'substatus': ContentStatus.New,
                         'status': ContentStatus.New}
            updated_contents.append(u_content)
    return updated_contents


def handle_resume_processing(processing, agent_attributes, logger=None, log_prefix=''):
    logger = get_logger(logger)

    request_id = processing['request_id']
    transform_id = processing['transform_id']
    workload_id = processing['workload_id']

    proc = processing['processing_metadata']['processing']
    work = proc.work
    work.set_agent_attributes(agent_attributes, processing)

    work.resume_processing(processing, log_prefix=log_prefix)

    input_collections = work.get_input_collections()
    output_collections = work.get_output_collections()
    log_collections = work.get_log_collections()

    update_collections = []
    for coll in input_collections + output_collections + log_collections:
        coll.status = CollectionStatus.Open
        coll.substatus = CollectionStatus.Open
        u_collection = {'coll_id': coll.coll_id,
                        'status': CollectionStatus.Open,
                        'substatus': CollectionStatus.Open}
        update_collections.append(u_collection)

    input_output_maps = get_input_output_maps(transform_id, work)
    update_contents = reactive_contents(request_id, transform_id, workload_id, work, input_output_maps)

    processing['status'] = ProcessingStatus.Running
    return processing, update_collections, update_contents



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


"""
Test active learning client.
"""

import os

from idds.client.client import Client
from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import get_rest_host


def get_test_codes():
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_codes = os.path.join(dir_name, 'activelearning_test_codes/activelearning_test_codes.tgz')
    return test_codes


def get_req_properties():
    req_properties = {
        'scope': 'data15_13TeV',
        'name': 'data15_13TeV.00270949.physics_Main.merge.AOD.r7600_p2521_tid07734829_00',
        'requester': 'panda',
        'request_type': RequestType.ActiveLearning,
        'transform_tag': 'prodsys2',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': '20525134', 'sandbox': 'https://', 'executable': 'hostname', 'arguments': '-s --input %IN', 'output_json': 'output.json'}
    }
    return req_properties


host = get_rest_host()
props = get_req_properties()
test_codes = get_test_codes()

client = Client(host=host)
test_codes_url = client.upload(test_codes)
props['request_metadata']['sandbox'] = test_codes_url
props['request_metadata']['executable'] = 'test.sh'
props['request_metadata']['arguments'] = '-1 -2 test'
# props['request_metadata']['result_parser'] = 'default'

request_id = client.add_request(**props)
print(request_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
operations related to collections.
"""

from idds.core import collections


def add_collection(scope, name, coll_type=None, request_id=None, transform_id=None,
                   relation_type=None, coll_size=0, status=None, total_files=0, retries=0,
                   expired_at=None, coll_metadata=None):
    """
    Add a collection.

    :param scope: The scope of the request data.
    :param name: The name of the request data.
    :param type: The type of dataset as dataset or container.
    :param request_id: The request id related to this collection.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.
    :param size: The size of the collection.
    :param status: The status.
    :param total_files: Number of total files.
    :param retries: Number of retries.
    :param expired_at: The datetime when it expires.
    :param coll_metadata: The metadata as json.

    :returns: collection id.
    """
    kwargs = {'scope': scope, 'name': name, 'coll_type': coll_type, 'request_id': request_id,
              'transform_id': transform_id, 'relation_type': relation_type, 'coll_size': coll_size,
              'status': status, 'total_files': total_files, 'retries': retries,
              'expired_at': expired_at, 'coll_metadata': coll_metadata}
    return collections.add_collection(**kwargs)


def get_collection(coll_id=None, transform_id=None, relation_type=None):
    """
    Get a collection or raise a NoObject exception.

    :param coll_id: The id of the collection.
    :param transform_id: The transform id related to this collection.
    :param relation_type: The relation between this collection and its transform,
                          such as Input, Output, Log and so on.

    :returns: Collection.
    """
    return collections.get_collection(coll_id=coll_id, transform_id=transform_id, relation_type=relation_type)


def update_collection(coll_id, parameters):
    """
    update a collection.

    :param coll_id: the collection id.
    :param parameters: A dictionary of parameters.

    """
    return collections.update_collection(coll_id=coll_id, parameters=parameters)


def delete_collection(coll_id=None):
    """
    delete a collection.

    :param request_id: The id of the request.

    """
    return collections.delete_collection(coll_id=coll_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020

import traceback
try:
    # python 3
    from queue import Queue
except ImportError:
    # Python 2
    from Queue import Queue


from idds.common.constants import (Sections, WorkprogressStatus, WorkprogressLocking, TransformStatus)
from idds.common.utils import setup_logging
from idds.core import (workprogress as core_workprogress, transforms as core_transforms)
from idds.workflow.work import WorkStatus
from idds.agents.common.baseagent import BaseAgent

setup_logging(__name__)


class Marshaller(BaseAgent):
    """
    Marshaller works to organize workflow, activate the current works in a workflow to workprogresses..
    """

    def __init__(self, num_threads=1, poll_time_period=1800, retrieve_bulk_size=10, **kwargs):
        super(Marshaller, self).__init__(num_threads=num_threads, **kwargs)
        self.config_section = Sections.Marshaller
        self.poll_time_period = int(poll_time_period)
        self.retrieve_bulk_size = int(retrieve_bulk_size)

        self.new_task_queue = Queue()
        self.new_output_queue = Queue()
        self.running_task_queue = Queue()
        self.running_output_queue = Queue()

    def get_new_workprogresses(self):
        """
        Get new workprogress to process
        """

        workprogress_status = [WorkprogressStatus.New, WorkprogressStatus.Ready, WorkprogressStatus.Extend]
        workprogresses_new = core_workprogress.get_workprogresses_by_status(status=workprogress_status,
                                                                            locking=True,
                                                                            bulk_size=self.retrieve_bulk_size)

        self.logger.debug("Main thread get %s New+Ready+Extend workprogresses to process" % len(workprogresses_new))
        if workprogresses_new:
            self.logger.info("Main thread get %s New+Ready+Extend workprogresses to process" % len(workprogresses_new))
        return workprogresses_new

    def process_new_workprogress(self, workprogress):
        """
        Process new workprogress
        """
        wf = workprogress['workprogress_metadata']['workflow']
        works = wf.get_new_works()

        transforms = []
        for work in works:
            new_work = work.copy()
            new_work.add_proxy(wf.get_proxy())
            transform = {'workprogress_id': workprogress['workprogress_id'],
                         'request_id': workprogress['request_id'],
                         'workload_id': workprogress['workload_id'],
                         'transform_type': work.get_work_type(),
                         'transform_tag': work.get_work_tag(),
                         'priority': workprogress['priority'],
                         'status': TransformStatus.New,
                         'retries': 0,
                         'expired_at': workprogress['expired_at'],
                         'transform_metadata': {'orginal_work': work, 'work': new_work}
                         # 'collections': related_collections
                         }
            transforms.append(transform)

        self.logger.info("Processing workprogress(%s): new transforms: %s" % (workprogress['workprogress_id'],
                                                                              transforms))

        workprogress['locking'] = WorkprogressLocking.Idle
        workprogress['status'] = WorkprogressStatus.Transforming
        ret_wp = {'workprogress': workprogress, 'new_transforms': transforms}
        return ret_wp

    def process_new_workprogresses(self):
        ret = []
        while not self.new_task_queue.empty():
            try:
                workprogress = self.new_task_queue.get()
                if workprogress:
                    self.logger.info("Main thread processing new workprogress: %s" % workprogress)
                    ret_workprogress = self.process_new_workprogress(workprogress)
                    if ret_workprogress:
                        ret.append(ret_workprogress)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return ret

    def finish_new_workprogresses(self):
        while not self.new_output_queue.empty():
            try:
                ret = self.new_output_queue.get()
                self.logger.info("Main thread finishing new workprogress: %s" % ret['workprogress'])
                if ret:
                    wp = ret['workprogress']
                    tfs = ret['new_transforms']
                    wp_parameters = {'status': wp['status'],
                                     'locking': wp['locking'],
                                     'workprogress_metadata': wp['workprogress_metadata']}
                    core_workprogress.update_workprogress(workprogress_id=wp['workprogress_id'],
                                                          parameters=wp_parameters,
                                                          new_transforms=tfs)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())

    def get_running_workprogresses(self):
        """
        Get workprogresses to running
        """
        workprogress_status = [WorkprogressStatus.Transforming, WorkprogressStatus.ToCancel,
                               WorkprogressStatus.Cancelling]
        workprogresses = core_workprogress.get_workprogresses_by_status(status=workprogress_status,
                                                                        period=self.poll_time_period,
                                                                        locking=True,
                                                                        bulk_size=self.retrieve_bulk_size)

        self.logger.debug("Main thread get %s progressing workprogresses to process" % len(workprogresses))
        if workprogresses:
            self.logger.info("Main thread get %s progressing workprogresses to process" % len(workprogresses))
        return workprogresses

    def process_running_workprogress(self, workprogress):
        """
        process running workprogresses
        """
        self.logger.info("process_running_workprogress: workprogress_id: %s" % workprogress['workprogress_id'])
        workprogress_metadata = workprogress['workprogress_metadata']
        wf = workprogress_metadata['workflow']

        new_transforms = []
        if workprogress['status'] in [WorkprogressStatus.Transforming]:
            # new works
            works = wf.get_new_works()
            for work in works:
                new_work = work.copy()
                new_work.add_proxy(wf.get_proxy())
                new_transform = {'workprogress_id': workprogress['workprogress_id'],
                                 'request_id': workprogress['request_id'],
                                 'workload_id': workprogress['workload_id'],
                                 'transform_type': work.get_work_type(),
                                 'transform_tag': work.get_work_tag(),
                                 'priority': workprogress['priority'],
                                 'status': TransformStatus.New,
                                 'retries': 0,
                                 'expired_at': workprogress['expired_at'],
                                 'transform_metadata': {'orginal_work': work, 'work': new_work}
                                 # 'collections': related_collections
                                 }
                new_transforms.append(new_transform)
            self.logger.info("Processing workprogress(%s): new transforms: %s" % (workprogress['workprogress_id'],
                                                                                  new_transforms))

        update_transforms = {}
        if workprogress['status'] in [WorkprogressStatus.ToCancel]:
            # current works
            works = wf.get_current_works()
            # print(works)
            for work in works:
                if work.get_status() not in [WorkStatus.Finished, WorkStatus.SubFinished,
                                             WorkStatus.Failed, WorkStatus.Cancelling,
                                             WorkStatus.Cancelled]:
                    update_transforms[work.get_work_id()] = {'status': TransformStatus.ToCancel}

        # current works
        works = wf.get_current_works()
        # print(works)
        for work in works:
            # print(work.get_work_id())
            tf = core_transforms.get_transform(transform_id=work.get_work_id())
            work_status = WorkStatus(tf['status'].value)
            work.set_status(work_status)
            work.set_terminated_msg(msg=None)   # TODO

        if wf.is_terminated():
            if wf.is_finished():
                wp_status = WorkprogressStatus.Finished
            elif wf.is_subfinished():
                wp_status = WorkprogressStatus.SubFinished
            elif wf.is_failed():
                wp_status = WorkprogressStatus.Failed
            elif wf.is_cancelled():
                wp_status = WorkprogressStatus.Cancelled
            else:
                wp_status = WorkprogressStatus.Failed
            wp_msg = wf.get_terminated_msg()
        else:
            wp_status = WorkprogressStatus.Transforming
            wp_msg = None
        parameters = {'status': wp_status,
                      'locking': WorkprogressLocking.Idle,
                      'workprogress_metadata': workprogress_metadata,
                      'errors': {'msg': wp_msg}}
        ret = {'workprogress_id': workprogress['workprogress_id'],
               'parameters': parameters,
               'new_transforms': new_transforms,
               'update_transforms': update_transforms}
        return ret

    def process_running_workprogresses(self):
        ret = []
        while not self.running_task_queue.empty():
            try:
                workprogress = self.running_task_queue.get()
                if workprogress:
                    self.logger.info("Main thread processing running workprogress: %s" % workprogress)
                    ret_workprogress = self.process_running_workprogress(workprogress)
                    if ret_workprogress:
                        ret.append(ret_workprogress)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return ret

    def finish_running_workprogresses(self):
        while not self.running_output_queue.empty():
            try:
                ret = self.running_output_queue.get()
                self.logger.info("Main thread finishing processing workprogress: %s" % ret)

                wp_id = ret['workprogress_id']
                parameters = ret['parameters']
                new_transforms = ret['new_transforms']
                update_transforms = ret['update_transforms']
                core_workprogress.update_workprogress(workprogress_id=wp_id,
                                                      parameters=parameters,
                                                      new_transforms=new_transforms,
                                                      update_transforms=update_transforms)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())

    def clean_locks(self):
        self.logger.info("clean locking")
        core_workprogress.clean_locking()

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")

            self.load_plugins()

            self.add_default_tasks()

            task = self.create_task(task_func=self.get_new_workprogresses, task_output_queue=self.new_task_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            for _ in range(self.num_threads):
                task = self.create_task(task_func=self.process_new_workprogresses, task_output_queue=self.new_output_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
                self.add_task(task)
            task = self.create_task(task_func=self.finish_new_workprogresses, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=2, priority=1)
            self.add_task(task)

            task = self.create_task(task_func=self.get_running_workprogresses, task_output_queue=self.running_task_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            for _ in range(self.num_threads):
                task = self.create_task(task_func=self.process_running_workprogresses, task_output_queue=self.running_output_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
                self.add_task(task)
            task = self.create_task(task_func=self.finish_running_workprogresses, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)

            task = self.create_task(task_func=self.clean_locks, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=1800, priority=1)
            self.add_task(task)

            self.execute()
        except KeyboardInterrupt:
            self.stop()


if __name__ == '__main__':
    agent = Marshaller()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019

import copy
import datetime
import traceback
try:
    # python 3
    from queue import Queue
except ImportError:
    # Python 2
    from Queue import Queue

from idds.common.constants import (Sections, CollectionRelationType, CollectionStatus,
                                   CollectionLocking, CollectionType, ContentType, ContentStatus,
                                   TransformType, ProcessingStatus,
                                   MessageType, MessageStatus, MessageSource)
from idds.common.exceptions import AgentPluginError
from idds.common.utils import setup_logging
from idds.core import (catalog as core_catalog, transforms as core_transforms, processings as core_processings)
from idds.agents.common.baseagent import BaseAgent

setup_logging(__name__)


class Transporter(BaseAgent):
    """
    Transporter works to list collections from DDM and register contents to DDM.
    """

    def __init__(self, num_threads=1, poll_time_period=10, retrieve_bulk_size=None, poll_input_time_period=None, poll_output_time_period=None, **kwargs):
        super(Transporter, self).__init__(num_threads=num_threads, **kwargs)
        self.poll_time_period = int(poll_time_period)
        if poll_input_time_period is None:
            self.poll_input_time_period = self.poll_time_period
        else:
            self.poll_input_time_period = int(poll_input_time_period)
        if poll_output_time_period is None:
            self.poll_output_time_period = self.poll_time_period
        else:
            self.poll_output_time_period = int(poll_output_time_period)

        self.retrieve_bulk_size = int(retrieve_bulk_size)
        self.config_section = Sections.Transporter

        self.new_input_queue = Queue()
        self.processed_input_queue = Queue()
        self.new_output_queue = Queue()
        self.processed_output_queue = Queue()

    def init(self):
        status = [CollectionStatus.New, CollectionStatus.Open,
                  CollectionStatus.Updated, CollectionStatus.Processing]
        core_catalog.clean_next_poll_at(status)

    def get_new_input_collections(self):
        """
        Get new input collections
        """
        coll_status = [CollectionStatus.Open]
        colls_open = core_catalog.get_collections_by_status(status=coll_status,
                                                            relation_type=CollectionRelationType.Input,
                                                            # time_period=self.poll_input_time_period,
                                                            locking=True,
                                                            bulk_size=self.retrieve_bulk_size)

        self.logger.debug("Main thread get %s [open] input collections to process" % len(colls_open))
        if colls_open:
            self.logger.info("Main thread get %s [open] input collections to process" % len(colls_open))

        coll_status = [CollectionStatus.New]
        colls_new = core_catalog.get_collections_by_status(status=coll_status,
                                                           relation_type=CollectionRelationType.Input,
                                                           locking=True,
                                                           bulk_size=self.retrieve_bulk_size)

        self.logger.debug("Main thread get %s [new] input collections to process" % len(colls_new))
        if colls_new:
            self.logger.info("Main thread get %s [new] input collections to process" % len(colls_new))

        colls = colls_open + colls_new
        self.logger.debug("Main thread get totally %s [open + new] collections to process" % len(colls))
        if colls:
            self.logger.info("Main thread get totally %s [open + new] collections to process" % len(colls))

        return colls

    def get_collection_metadata(self, scope, name):
        if 'collection_metadata_reader' not in self.plugins:
            raise AgentPluginError('Plugin collection_metadata_reader is required')
        return self.plugins['collection_metadata_reader'](scope, name)

    def get_contents(self, scope, name):
        if 'contents_lister' not in self.plugins:
            raise AgentPluginError('Plugin contents_lister is required')
        return self.plugins['contents_lister'](scope, name)

    def process_input_collection(self, coll):
        """
        Process input collection
        """
        if coll['coll_type'] in [CollectionType.PseudoDataset, CollectionType.PseudoDataset.value]:
            pseudo_content = {'coll_id': coll['coll_id'],
                              'scope': coll['scope'],
                              'name': 'pseudo_%s' % coll['coll_id'],
                              'min_id': 0,
                              'max_id': 0,
                              'content_type': ContentType.PseudoContent,
                              'status': ContentStatus.Available,
                              'bytes': 0,
                              'md5': None,
                              'adler32': None,
                              'expired_at': coll['expired_at']}
            new_coll = {'coll': coll,
                        'bytes': 0,
                        'status': CollectionStatus.Open,
                        'total_files': 1,
                        'new_files': 0,
                        'processed_files': 0,
                        'coll_metadata': {'availability': True,
                                          'events': 0,
                                          'is_open': False,
                                          'status': CollectionStatus.Closed.name},
                        'contents': [pseudo_content]}
        elif coll['coll_metadata'] and coll['coll_metadata']['status'] in [CollectionStatus.Closed, CollectionStatus.Closed.name, CollectionStatus.Closed.value]:
            new_coll = {'coll': coll, 'status': coll['status'], 'contents': []}
        else:
            coll_metadata = self.get_collection_metadata(coll['scope'], coll['name'])
            contents = self.get_contents(coll['scope'], coll['name'])
            new_contents = []
            for content in contents:
                new_content = {'coll_id': coll['coll_id'],
                               'scope': content['scope'],
                               'name': content['name'],
                               'min_id': 0,
                               'max_id': content['events'],
                               'content_type': ContentType.File,
                               'status': ContentStatus.Available,
                               'bytes': content['bytes'],
                               'md5': content['md5'] if 'md5' in content else None,
                               'adler32': content['adler32'] if 'adler32' in content else None,
                               'expired_at': coll['expired_at']}
                new_contents.append(new_content)

            coll['bytes'] = coll_metadata['bytes']
            coll['total_files'] = coll_metadata['total_files']
            new_coll = {'coll': coll,
                        'bytes': coll_metadata['bytes'],
                        'status': CollectionStatus.Open,
                        'total_files': coll_metadata['total_files'],
                        'coll_metadata': {'availability': coll_metadata['availability'],
                                          'events': coll_metadata['events'],
                                          'is_open': coll_metadata['is_open'],
                                          'status': coll_metadata['status'].name,
                                          'run_number': coll_metadata['run_number']},
                        'contents': new_contents}
        return new_coll

    def process_input_collections(self):
        ret = []
        while not self.new_input_queue.empty():
            try:
                coll = self.new_input_queue.get()
                if coll:
                    self.logger.info("Main thread processing input collection: %s" % coll)
                    ret_coll = self.process_input_collection(coll)
                    if ret_coll:
                        ret.append(ret_coll)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return ret

    def finish_processing_input_collections(self):
        while not self.processed_input_queue.empty():
            coll = self.processed_input_queue.get()
            self.logger.info("Main thread finished processing intput collection(%s) with number of contents: %s" % (coll['coll']['coll_id'], len(coll['contents'])))
            parameters = copy.deepcopy(coll)
            if parameters['coll']['coll_type'] in [CollectionType.PseudoDataset, CollectionType.PseudoDataset.value]:
                pass
            else:
                parameters['next_poll_at'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.poll_input_time_period)
            del parameters['contents']
            del parameters['coll']
            parameters['locking'] = CollectionLocking.Idle
            core_catalog.update_input_collection_with_contents(coll=coll['coll'],
                                                               parameters=parameters,
                                                               contents=coll['contents'])

    def get_new_output_collections(self):
        """
        Get new output collections
        """
        coll_status = [CollectionStatus.Updated, CollectionStatus.Processing]
        colls = core_catalog.get_collections_by_status(status=coll_status,
                                                       relation_type=CollectionRelationType.Output,
                                                       # time_period=self.poll_output_time_period,
                                                       locking=True,
                                                       bulk_size=self.retrieve_bulk_size)
        self.logger.debug("Main thread get %s [Updated + Processing] output collections to process" % len(colls))
        if colls:
            self.logger.info("Main thread get %s [Updated + Processing] output collections to process" % len(colls))
        return colls

    def register_contents(self, scope, name, contents):
        if 'contents_register' not in self.plugins:
            raise AgentPluginError('Plugin contents_register is required')
        return self.plugins['contents_register'](scope, name, contents)

    def is_input_collection_all_processed(self, coll_id_list):
        is_all_processed = True
        for coll_id in coll_id_list:
            coll = core_catalog.get_collection(coll_id)
            if not (coll['status'] == CollectionStatus.Closed and coll['total_files'] == coll['processed_files']):
                is_all_processed = False
                return is_all_processed
        return is_all_processed

    def is_all_processings_finished(self, transform_id):
        last_processing = None
        processings = core_processings.get_processings_by_transform_id(transform_id=transform_id)
        if processings:
            last_processing = processings[-1]

        for processing in processings:
            if not processing['status'] in [ProcessingStatus.Finished, ProcessingStatus.Failed, ProcessingStatus.Lost,
                                            # ProcessingStatus.Cancel, ProcessingStatus.FinishedOnStep, ProcessingStatus.FinishedOnExec,
                                            ProcessingStatus.Cancel, ProcessingStatus.FinishedOnStep,
                                            ProcessingStatus.TimeOut,
                                            ProcessingStatus.Finished.value, ProcessingStatus.Failed.value, ProcessingStatus.Lost.value,
                                            # ProcessingStatus.Cancel.value, ProcessingStatus.FinishedOnStep.value, ProcessingStatus.FinishedOnExec.value,
                                            ProcessingStatus.Cancel.value, ProcessingStatus.FinishedOnStep.value,
                                            ProcessingStatus.TimeOut.value]:
                return False, last_processing
        return True, last_processing

    def process_output_collection(self, coll):
        """
        Process output collection
        """
        if coll['coll_metadata'] and 'to_register' in coll['coll_metadata'] and coll['coll_metadata']['to_register'] is True:
            contents = core_catalog.get_contents(coll_id=coll['coll_id'])
            registered_contents = self.get_contents(coll['scope'], coll['name'])
            registered_content_names = [con['name'] for con in registered_contents]
            new_contents = []
            for content in contents:
                if content['name'] not in registered_content_names:
                    new_contents.append(content)
            if new_contents:
                self.register_contents(coll['scope'], coll['name'], new_contents)

        is_all_processings_finished, last_processing = self.is_all_processings_finished(coll['transform_id'])

        is_input_collection_all_processed = False
        if coll['coll_metadata'] and 'input_collections' in coll['coll_metadata'] and coll['coll_metadata']['input_collections']:
            input_coll_list = coll['coll_metadata']['input_collections']
            is_input_collection_all_processed = self.is_input_collection_all_processed(input_coll_list)

        contents_statistics = core_catalog.get_content_status_statistics(coll_id=coll['coll_id'])
        contents_statistics_with_name = {}
        for key in contents_statistics:
            contents_statistics_with_name[key.name] = contents_statistics[key]

        content_status_keys = list(contents_statistics.keys())
        total_files = sum(contents_statistics.values())
        new_files = 0
        processed_files = 0
        # output_metadata = None
        if ContentStatus.Available in contents_statistics:
            processed_files += contents_statistics[ContentStatus.Available]
        if ContentStatus.Available.value in contents_statistics:
            processed_files += contents_statistics[ContentStatus.Available.value]
        if ContentStatus.New in contents_statistics:
            new_files += contents_statistics[ContentStatus.New]
        if ContentStatus.New.value in contents_statistics:
            new_files += contents_statistics[ContentStatus.New.value]

        coll_msg = None
        if not is_input_collection_all_processed or not is_all_processings_finished:
            coll_status = CollectionStatus.Processing
        elif content_status_keys == [ContentStatus.Available] or content_status_keys == [ContentStatus.Available.value]:
            coll_status = CollectionStatus.Closed

            transform = core_transforms.get_transform(coll['coll_metadata']['transform_id'])
            # if 'processing_id' in transform['transform_metadata']:
            #    processing = core_processings.get_processing(transform['transform_metadata']['processing_id'])
            #    output_metadata = processing['output_metadata']
            output_ret = {}
            if last_processing:
                output_ret = {'status': last_processing['status'].name,
                              'msg': last_processing['processing_metadata']['final_errors'] if 'final_errors' in last_processing['processing_metadata'] else None}
            if transform['transform_type'] in [TransformType.StageIn, TransformType.StageIn.value]:
                msg_type = 'collection_stagein'
                msg_type_c = MessageType.StageInCollection
            elif transform['transform_type'] in [TransformType.ActiveLearning, TransformType.ActiveLearning.value]:
                msg_type = 'collection_activelearning'
                msg_type_c = MessageType.ActiveLearningCollection
            elif transform['transform_type'] in [TransformType.HyperParameterOpt, TransformType.HyperParameterOpt.value]:
                msg_type = 'collection_hyperparameteropt'
                msg_type_c = MessageType.HyperParameterOptCollection
            else:
                msg_type = 'collection_unknown'
                msg_type_c = MessageType.UnknownCollection

            msg_content = {'msg_type': msg_type,
                           'workload_id': coll['coll_metadata']['workload_id'] if 'workload_id' in coll['coll_metadata'] else None,
                           'collections': [{'scope': coll['scope'],
                                            'name': coll['name'],
                                            'status': 'Available'}],
                           'output': output_ret}
            coll_msg = {'msg_type': msg_type_c,
                        'status': MessageStatus.New,
                        'source': MessageSource.Transporter,
                        'transform_id': coll['coll_metadata']['transform_id'] if 'transform_id' in coll['coll_metadata'] else None,
                        'num_contents': 1,
                        'msg_content': msg_content}

        elif content_status_keys == [ContentStatus.FinalFailed] or content_status_keys == [ContentStatus.FinalFailed.value]:
            coll_status = CollectionStatus.Failed
        elif (len(content_status_keys) == 2                                                                                   # noqa: W503
            and (ContentStatus.FinalFailed in content_status_keys or ContentStatus.FinalFailed.value in content_status_keys)  # noqa: W503, E128
            and (ContentStatus.Available in content_status_keys or ContentStatus.Available.value in content_status_keys)):    # noqa: W503, E128, E125
            coll_status = CollectionStatus.SubClosed
        elif (ContentStatus.New in content_status_keys or ContentStatus.New.value in content_status_keys            # noqa: W503
            or ContentStatus.Failed in content_status_keys or ContentStatus.Failed.value in content_status_keys):   # noqa: W503, E128, E125
            # coll_status = CollectionStatus.Processing
            coll_status = CollectionStatus.Failed
        else:
            # coll_status = CollectionStatus.Processing
            coll_status = CollectionStatus.Failed

        coll_metadata = coll['coll_metadata']
        if not coll_metadata:
            coll_metadata = {}
        coll_metadata['status_statistics'] = contents_statistics_with_name
        # coll_metadata['output_metadata'] = output_metadata
        ret_coll = {'coll_id': coll['coll_id'],
                    'total_files': total_files,
                    'status': coll_status,
                    'new_files': new_files,
                    'processing_files': 0,
                    'processed_files': processed_files,
                    'coll_metadata': coll_metadata,
                    'coll_msg': coll_msg}

        if coll['coll_type'] in [CollectionType.PseudoDataset, CollectionType.PseudoDataset.value]:
            pass
        else:
            ret_coll['next_poll_at'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.poll_output_time_period)

        return ret_coll

    def process_output_collections(self):
        ret = []
        while not self.new_output_queue.empty():
            try:
                coll = self.new_output_queue.get()
                if coll:
                    self.logger.info("Main thread processing output collection: %s" % coll)
                    ret_coll = self.process_output_collection(coll)
                    if ret_coll:
                        ret.append(ret_coll)
            except Exception as ex:
                self.logger.error(ex)
                self.logger.error(traceback.format_exc())
        return ret

    def finish_processing_output_collections(self):
        while not self.processed_output_queue.empty():
            coll = self.processed_output_queue.get()
            self.logger.info("Main thread finished processing output collection(%s) with number of processed contents: %s" % (coll['coll_id'], coll['processed_files']))
            coll_id = coll['coll_id']
            coll_msg = coll['coll_msg']
            parameters = coll
            del parameters['coll_id']
            del coll['coll_msg']
            parameters['locking'] = CollectionLocking.Idle
            core_catalog.update_collection(coll_id=coll_id, parameters=parameters, msg=coll_msg)

    def clean_locks(self):
        self.logger.info("clean locking")
        core_catalog.clean_locking()

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")

            self.load_plugins()
            self.init()

            self.add_default_tasks()

            task = self.create_task(task_func=self.get_new_input_collections, task_output_queue=self.new_input_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            task = self.create_task(task_func=self.process_input_collections, task_output_queue=self.processed_input_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            task = self.create_task(task_func=self.finish_processing_input_collections, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)

            task = self.create_task(task_func=self.get_new_output_collections, task_output_queue=self.new_output_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            task = self.create_task(task_func=self.process_output_collections, task_output_queue=self.processed_output_queue, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)
            task = self.create_task(task_func=self.finish_processing_output_collections, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=1, priority=1)
            self.add_task(task)

            task = self.create_task(task_func=self.clean_locks, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=1800, priority=1)
            self.add_task(task)

            self.execute()
        except KeyboardInterrupt:
            self.stop()


if __name__ == '__main__':
    agent = Transporter()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019 - 2022

import traceback

from idds.common.constants import (Sections, ProcessingStatus, ProcessingLocking)
from idds.common.utils import setup_logging, truncate_string
from idds.agents.common.eventbus.event import (EventType,
                                               UpdateProcessingEvent,
                                               UpdateTransformEvent)

from .utils import (handle_abort_processing,
                    handle_resume_processing,
                    # is_process_terminated,
                    sync_processing)
from .poller import Poller

setup_logging(__name__)


class Finisher(Poller):
    """
    Finisher works to submit and running tasks to WFMS.
    """

    def __init__(self, num_threads=1, poll_time_period=10, retries=3, retrieve_bulk_size=2,
                 message_bulk_size=1000, **kwargs):
        self.set_max_workers()
        if hasattr(self, 'finisher_max_number_workers'):
            self.max_number_workers = int(self.finisher_max_number_workers)
        num_threads = self.max_number_workers

        super(Finisher, self).__init__(num_threads=num_threads, name='Finisher',
                                       poll_time_period=poll_time_period, retries=retries,
                                       retrieve_bulk_size=retrieve_bulk_size,
                                       message_bulk_size=message_bulk_size, **kwargs)
        self.config_section = Sections.Carrier
        self.poll_time_period = int(poll_time_period)
        self.retries = int(retries)

        if hasattr(self, 'finisher_max_number_workers'):
            self.max_number_workers = int(self.finisher_max_number_workers)

    def show_queue_size(self):
        q_str = "number of processings: %s, max number of processings: %s" % (self.number_workers, self.max_number_workers)
        self.logger.debug(q_str)

    def handle_sync_processing(self, processing, log_prefix=""):
        """
        process terminated processing
        """
        try:
            processing, update_collections, messages = sync_processing(processing, self.agent_attributes, logger=self.logger, log_prefix=log_prefix)

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing['status'],
                                                'locking': ProcessingLocking.Idle}}
            ret = {'update_processing': update_processing,
                   'update_collections': update_collections,
                   'messages': messages}
            return ret
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            error = {'sync_err': {'msg': truncate_string('%s' % (ex), length=200)}}
            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': ProcessingStatus.Running,
                                                'locking': ProcessingLocking.Idle,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)
            ret = {'update_processing': update_processing}
            return ret
        return None

    def process_sync_processing(self, event):
        self.number_workers += 1
        try:
            if event:
                self.logger.info("process_sync_processing: event: %s" % event)
                pr = self.get_processing(processing_id=event._processing_id, locking=True)
                if not pr:
                    self.logger.error("Cannot find processing for event: %s" % str(event))
                else:
                    log_pre = self.get_log_prefix(pr)

                    self.logger.info(log_pre + "process_sync_processing")
                    ret = self.handle_sync_processing(pr, log_prefix=log_pre)
                    self.logger.info(log_pre + "process_sync_processing result: %s" % str(ret))

                    self.update_processing(ret, pr)

                    # no need to update transform
                    # self.logger.info(log_pre + "UpdateTransformEvent(transform_id: %s)" % pr['transform_id'])
                    # event = UpdateTransformEvent(publisher_id=self.id, transform_id=pr['transform_id'])
                    # self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
        self.number_workers -= 1

    def handle_terminated_processing(self, processing, log_prefix=""):
        """
        process terminated processing
        """
        try:
            processing, update_collections, messages = sync_processing(processing, self.agent_attributes, terminate=True, logger=self.logger, log_prefix=log_prefix)

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing['status'],
                                                'locking': ProcessingLocking.Idle}}
            ret = {'update_processing': update_processing,
                   'update_collections': update_collections,
                   'messages': messages}

            return ret
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            error = {'term_err': {'msg': truncate_string('%s' % (ex), length=200)}}
            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': ProcessingStatus.Running,
                                                'locking': ProcessingLocking.Idle,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)
            ret = {'update_processing': update_processing}
            return ret
        return None

    def process_terminated_processing(self, event):
        self.number_workers += 1
        try:
            if event:
                if event._counter > 3:
                    self.logger.warn("Event counter is bigger than 3, skip event: %s" % str(event))
                else:
                    original_event = event
                    pr = self.get_processing(processing_id=event._processing_id, locking=True)
                    if not pr:
                        self.logger.error("Cannot find processing for event: %s" % str(event))
                    else:
                        log_pre = self.get_log_prefix(pr)

                        self.logger.info(log_pre + "process_terminated_processing")
                        ret = self.handle_terminated_processing(pr, log_prefix=log_pre)
                        self.logger.info(log_pre + "process_terminated_processing result: %s" % str(ret))

                        self.update_processing(ret, pr)
                        self.logger.info(log_pre + "UpdateTransformEvent(transform_id: %s)" % pr['transform_id'])
                        event = UpdateTransformEvent(publisher_id=self.id, transform_id=pr['transform_id'])
                        self.event_bus.send(event)

                        if pr['status'] not in [ProcessingStatus.Finished, ProcessingStatus.Failed, ProcessingStatus.SubFinished]:
                            # some files are missing, poll it.
                            self.logger.info(log_pre + "UpdateProcessingEvent(processing_id: %s)" % pr['processing_id'])
                            event = UpdateProcessingEvent(publisher_id=self.id, processing_id=pr['processing_id'], counter=original_event._counter + 1)
                            self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
        self.number_workers -= 1

    def handle_abort_processing(self, processing, log_prefix=""):
        """
        process abort processing
        """
        try:
            processing, update_collections, update_contents, messages = handle_abort_processing(processing, self.agent_attributes, logger=self.logger, log_prefix=log_prefix)

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing['status'],
                                                'locking': ProcessingLocking.Idle}}
            ret = {'update_processing': update_processing,
                   'update_collections': update_collections,
                   'update_contents': update_contents,
                   'messages': messages
                   }
            return ret
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            error = {'abort_err': {'msg': truncate_string('%s' % (ex), length=200)}}
            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': ProcessingStatus.ToCancel,
                                                'locking': ProcessingLocking.Idle,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)
            ret = {'update_processing': update_processing}
            return ret
        return None

    def process_abort_processing(self, event):
        self.number_workers += 1
        try:
            if event:
                processing_status = [ProcessingStatus.Finished, ProcessingStatus.Failed,
                                     ProcessingStatus.Lost, ProcessingStatus.Cancelled,
                                     ProcessingStatus.Suspended, ProcessingStatus.Expired,
                                     ProcessingStatus.Broken]

                pr = self.get_processing(processing_id=event._processing_id, locking=True)

                if not pr:
                    self.logger.error("Cannot find processing for event: %s" % str(event))
                else:
                    log_pre = self.get_log_prefix(pr)
                    self.logger.info(log_pre + "process_abort_processing")

                    if pr and pr['status'] in processing_status:
                        update_processing = {'processing_id': pr['processing_id'],
                                             'parameters': {'locking': ProcessingLocking.Idle,
                                                            'errors': {'abort_err': {'msg': truncate_string("Processing is already terminated. Cannot be aborted", length=200)}}}}
                        ret = {'update_processing': update_processing}
                        self.logger.info(log_pre + "process_abort_processing result: %s" % str(ret))
                        self.update_processing(ret, pr)
                    elif pr:
                        ret = self.handle_abort_processing(pr, log_prefix=log_pre)
                        self.logger.info(log_pre + "process_abort_processing result: %s" % str(ret))
                        self.update_processing(ret, pr)
                        self.logger.info(log_pre + "UpdateTransformEvent(transform_id: %s)" % pr['transform_id'])
                        event = UpdateTransformEvent(publisher_id=self.id, transform_id=pr['transform_id'], content=event._content)
                        self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
        self.number_workers -= 1

    def handle_resume_processing(self, processing, log_prefix=""):
        """
        process resume processing
        """
        try:
            processing, update_collections, update_contents = handle_resume_processing(processing, self.agent_attributes, logger=self.logger, log_prefix=log_prefix)

            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': processing['status'],
                                                'locking': ProcessingLocking.Idle}}
            ret = {'update_processing': update_processing,
                   'update_collections': update_collections,
                   'update_contents': update_contents,
                   }
            return ret
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
            error = {'resume_err': {'msg': truncate_string('%s' % (ex), length=200)}}
            update_processing = {'processing_id': processing['processing_id'],
                                 'parameters': {'status': ProcessingStatus.ToResume,
                                                'locking': ProcessingLocking.Idle,
                                                'errors': processing['errors'] if processing['errors'] else {}}}
            update_processing['parameters']['errors'].update(error)
            ret = {'update_processing': update_processing}
            return ret
        return None

    def process_resume_processing(self, event):
        self.number_workers += 1
        try:
            if event:
                processing_status = [ProcessingStatus.Finished]

                pr = self.get_processing(processing_id=event._processing_id, locking=True)

                if not pr:
                    self.logger.error("Cannot find processing for event: %s" % str(event))
                else:
                    log_pre = self.get_log_prefix(pr)
                    self.logger.info(log_pre + "process_resume_processing")

                    if pr and pr['status'] in processing_status:
                        update_processing = {'processing_id': pr['processing_id'],
                                             'parameters': {'locking': ProcessingLocking.Idle,
                                                            'errors': {'abort_err': {'msg': truncate_string("Processing has already finished. Cannot be resumed", length=200)}}}}
                        ret = {'update_processing': update_processing}

                        self.logger.info(log_pre + "process_resume_processing result: %s" % str(ret))

                        self.update_processing(ret, pr)
                    elif pr:
                        ret = self.handle_resume_processing(pr, log_prefix=log_pre)
                        self.logger.info(log_pre + "process_resume_processing result: %s" % str(ret))

                        self.update_processing(ret, pr)

                        self.logger.info(log_pre + "UpdateTransformEvent(transform_id: %s)" % pr['transform_id'])
                        event = UpdateTransformEvent(publisher_id=self.id, transform_id=pr['transform_id'], content=event._content)
                        self.event_bus.send(event)
        except Exception as ex:
            self.logger.error(ex)
            self.logger.error(traceback.format_exc())
        self.number_workers -= 1

    def init_event_function_map(self):
        self.event_func_map = {
            EventType.SyncProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_sync_processing
            },
            EventType.TerminatedProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_terminated_processing
            },
            EventType.AbortProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_abort_processing
            },
            EventType.ResumeProcessing: {
                'pre_check': self.is_ok_to_run_more_processings,
                'exec_func': self.process_resume_processing
            }
        }

    def run(self):
        """
        Main run function.
        """
        try:
            self.logger.info("Starting main thread")

            self.load_plugins()
            self.init()

            self.add_default_tasks()

            self.init_event_function_map()

            self.execute()
        except KeyboardInterrupt:
            self.stop()


if __name__ == '__main__':
    agent = Finisher()
    agent()



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019


"""
Test Request.
"""

import random
import time
import unittest2 as unittest
from nose.tools import assert_equal

from idds.common.utils import check_database, has_config, setup_logging
from idds.orm.requests import (add_request, get_request, delete_requests)
from idds.orm.transforms import (add_transform, get_transform, get_transform_ids, delete_transform)
from idds.tests.common import get_request_properties, get_transform_properties

setup_logging(__name__)


class TestTransform(unittest.TestCase):

    @unittest.skipIf(not has_config(), "No config file")
    @unittest.skipIf(not check_database(), "Database is not defined")
    def test_create_and_check_for_request_transform_orm(self):
        """ Transform (ORM): Test to create and delete a Transform """
        req_properties = get_request_properties()
        req_properties['workload_id'] = int(time.time()) + random.randint(1, 1000000)
        trans_properties = get_transform_properties()

        request_id = add_request(**req_properties)

        trans_properties['request_id'] = request_id
        trans_id = add_transform(**trans_properties)
        transform = get_transform(transform_id=trans_id)
        for key in trans_properties:
            if key in ['request_id']:
                continue
            assert_equal(transform[key], trans_properties[key])

        trans_ids1 = get_transform_ids(request_id=request_id)
        trans_ids2 = get_transform_ids(workload_id=req_properties['workload_id'])
        trans_ids3 = get_transform_ids(transform_id=trans_id)
        trans_ids4 = get_transform_ids(request_id=request_id, workload_id=req_properties['workload_id'])
        trans_ids5 = get_transform_ids(request_id=request_id, workload_id=req_properties['workload_id'], transform_id=trans_id)
        trans_ids6 = get_transform_ids(request_id=request_id, transform_id=trans_id)
        trans_ids7 = get_transform_ids(workload_id=req_properties['workload_id'], transform_id=trans_id)
        assert_equal(trans_ids1, trans_ids2)
        assert_equal(trans_ids2, trans_ids3)
        assert_equal(trans_ids3, trans_ids4)
        assert_equal(trans_ids4, trans_ids5)
        assert_equal(trans_ids5, trans_ids6)
        assert_equal(trans_ids6, trans_ids7)

        delete_transform(trans_id)
        delete_requests(request_id=request_id)

        req = get_request(request_id=request_id)
        assert_equal(req, None)

        trans = get_transform(transform_id=trans_id)
        assert_equal(trans, None)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020


import time
import traceback


class TimerTask(object):
    """
    The base class for Task which will be executed after some time
    """

    def __init__(self, task_func, task_output_queue=None, task_args=tuple(), task_kwargs={}, delay_time=10, priority=1, logger=None):
        self.to_execute_time = time.time()
        self.delay_time = delay_time
        self.priority = priority
        self.task_func = task_func
        self.task_output_queue = task_output_queue
        self.task_args = task_args
        self.task_kwargs = task_kwargs

        self.logger = logger

    def __eq__(one, two):
        return (one.to_execute_time, one.priority) == (two.to_execute_time, two.priority)

    def __lt__(one, two):
        return (one.to_execute_time, one.priority) < (two.to_execute_time, two.priority)

    def __le__(one, two):
        return (one.to_execute_time, one.priority) <= (two.to_execute_time, two.priority)

    def __gt__(one, two):
        return (one.to_execute_time, one.priority) > (two.to_execute_time, two.priority)

    def __ge__(one, two):
        return (one.to_execute_time, one.priority) >= (two.to_execute_time, two.priority)

    def is_ready(self):
        if self.to_execute_time <= time.time():
            return True
        return False

    def execute(self):
        try:
            # set it to avoid an exception
            self.to_execute_time = time.time() + self.delay_time

            ret = self.task_func(*self.task_args, **self.task_kwargs)
            if self.task_output_queue and ret is not None:
                for ret_item in ret:
                    self.task_output_queue.put(ret_item)

            # if there is no exception, this one is the correct one.
            self.to_execute_time = time.time() + self.delay_time
        except:
            if self.logger:
                self.logger.error('Failed to execute task func: %s, %s' % (self.task_func, traceback.format_exc()))
            else:
                print('Failed to execute task func: %s, %s' % (self.task_func, traceback.format_exc()))



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2019

"""
common funcs for tests
"""

import random
import time
import datetime
from uuid import uuid4 as uuid

from idds.common.constants import (RequestType, RequestStatus,
                                   TransformType, TransformStatus, CollectionType,
                                   CollectionRelationType, CollectionStatus,
                                   ContentType, ContentStatus,
                                   ProcessingStatus, GranularityType)


def get_request_properties():
    req_properties = {
        'scope': 'test_scope',
        'name': 'test_name_%s' % str(uuid()),
        'requester': 'panda',
        'request_type': RequestType.EventStreaming,
        'transform_tag': 's3218',
        'workload_id': int(time.time()) + random.randint(0, 1000000),
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id1': int(time.time())}
    }
    return req_properties


def get_transform_properties():
    trans_properties = {
        'transform_type': TransformType.EventStreaming,
        'transform_tag': 's3128',
        'priority': 0,
        'status': TransformStatus.New,
        'retries': 0,
        'expired_at': datetime.datetime.utcnow().replace(microsecond=0),
        'transform_metadata': {'input': {'coll_id': 123},
                               'output': {'coll_id': 456},
                               'log': {'coll_id': 789}}
    }
    return trans_properties


def get_collection_properties():
    coll_properties = {
        'scope': 'test_scope',
        'name': 'test_name_%s' % str(uuid()),
        'coll_type': CollectionType.Dataset,
        'transform_id': None,
        'relation_type': CollectionRelationType.Input,
        'bytes': 0,
        'status': CollectionStatus.New,
        'total_files': 0,
        'retries': 0,
        'expired_at': datetime.datetime.utcnow().replace(microsecond=0),
        'coll_metadata': {'ddm_status': 'closed'}
    }
    return coll_properties


def get_content_properties():
    content_properties = {
        'coll_id': None,
        'scope': 'test_scope',
        'name': 'test_file_name_%s' % str(uuid()),
        'min_id': 0,
        'max_id': 100,
        'content_type': ContentType.File,
        'status': ContentStatus.New,
        'bytes': 1,
        'md5': None,
        'adler32': None,
        'processing_id': None,
        'storage_id': None,
        'retries': 0,
        'path': None,
        'expired_at': datetime.datetime.utcnow().replace(microsecond=0),
        'content_metadata': {'id': 123}
    }
    return content_properties


def get_processing_properties():
    proc_properties = {
        'transform_id': None,
        'status': ProcessingStatus.New,
        'submitter': 'panda',
        'granularity': 10,
        'granularity_type': GranularityType.File,
        'expired_at': datetime.datetime.utcnow().replace(microsecond=0),
        'processing_metadata': {'task_id': 40191323}
    }
    return proc_properties


def get_example_tape_stagein_request():
    req_properties = {
        'scope': 'test_scope',
        'name': 'test_name_%s' % str(uuid()),
        'requester': 'panda',
        'request_type': RequestType.StageIn,
        'transform_tag': RequestType.StageIn.value,
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': int(time.time()), 'src_rse': 'BNL-OSG2_DATATAPE', 'dest_rse': 'BNL-OSG2_DATADISK'}
    }
    return req_properties


def get_example_real_tape_stagein_request():
    req_properties = {
        'scope': 'data17_13TeV',
        'name': 'data17_13TeV.00341649.express_express.merge.RAW',
        'requester': 'panda',
        'request_type': RequestType.StageIn,
        'transform_tag': RequestType.StageIn.value,
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': int(time.time()), 'src_rse': 'BNL-OSG2_DATATAPE', 'dest_rse': 'BNL-OSG2_DATADISK'}
    }
    return req_properties


def get_example_prodsys2_tape_stagein_request_old():
    req_properties = {
        'scope': 'data16_13TeV',
        'name': 'data16_13TeV.00298773.physics_Main.daq.RAW',
        'requester': 'panda',
        'request_type': RequestType.StageIn,
        'transform_tag': 'prodsys2',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': '20776834', 'src_rse': 'INFN-T1_DATATAPE', 'rule_id': 'ececf892223142b1b7c25b7593f50fe1'}
    }
    return req_properties


def get_example_prodsys2_tape_stagein_request():
    req_properties = {
        'scope': 'data16_13TeV',
        'name': 'data16_13TeV.00301915.physics_Main.daq.RAW',
        'requester': 'panda',
        'request_type': RequestType.StageIn,
        'transform_tag': 'prodsys2',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': '20776936', 'src_rse': 'NDGF-T1_DATATAPE', 'rule_id': '1d522d631327478db5773b554f65b0ff'}
    }
    return req_properties


def get_example_active_learning_request():
    req_properties = {
        'scope': 'data15_13TeV',
        'name': 'data15_13TeV.00270949.physics_Main.merge.AOD.r7600_p2521_tid07734829_00',
        'requester': 'panda',
        'request_type': RequestType.ActiveLearning,
        'transform_tag': 'prodsys2',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': '20525134', 'sandbox': 'https://', 'executable': 'hostname', 'arguments': '-s'}
    }
    return req_properties


def get_example_ess_request():
    req_properties = {
        'scope': 'test_scope',
        'name': 'test_name_%s' % str(uuid()),
        'requester': 'panda',
        'request_type': RequestType.EventStreaming,
        'transform_tag': 's3128',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': int(time.time()), 'events_per_range': 10}
    }
    return req_properties


# https://bigpanda.cern.ch/tasknew/19163542/
def get_real_example_ess_request():
    req_properties = {
        'scope': 'mc16_13TeV',
        'name': 'mc16_13TeV.451129.MGPy8EG_A14N23LO_VBF_RS_G_ZZ_llqq_kt1_m1500.merge.EVNT.e7758_e5984_tid19153490_00',
        'requester': 'panda',
        'request_type': RequestType.EventStreaming,
        'transform_tag': 's3126',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': 19163542, 'events_per_range': 10}
    }
    return req_properties


# https://bigpanda.cern.ch/task/19660169/
def get_example_derivation_request():
    req_properties = {
        'scope': 'mc16_13TeV',
        'name': 'mc16_13TeV.410287.PhPy8EG_A14_ttbar_hdamp258p75_allhad_mtt_1700_2000.merge.AOD.e6686_e5984_s3126_r10724_r10726_tid15801996_00',
        'requester': 'panda',
        'request_type': RequestType.Derivation,
        'transform_tag': 'p3978',
        'status': RequestStatus.New,
        'priority': 0,
        'lifetime': 30,
        'request_metadata': {'workload_id': 19660169, 'task_parameter': None}
    }
    return req_properties


def merge_dicts(list_dicts):
    ret = {}
    for a_dict in list_dicts:
        ret.update(a_dict)
    return ret


def is_same_req_trans_colls(req_trans_colls1, req_trans_colls2, allow_request_id_None=False):
    req_ids1 = list(req_trans_colls1.keys())
    req_ids2 = list(req_trans_colls2.keys())
    if not allow_request_id_None and not (req_ids1 == req_ids2):
        return False

    if not allow_request_id_None:
        for req_id in req_ids1:
            tran_ids1 = list(req_trans_colls1[req_id].keys())
            tran_ids2 = list(req_trans_colls2[req_id].keys())
            if not tran_ids1 == tran_ids2:
                return False

            for tran_id in tran_ids1:
                colls1 = req_trans_colls1[req_id][tran_id]
                colls2 = req_trans_colls2[req_id][tran_id]
                coll_ids1 = [coll['coll_id'] for coll in colls1]
                coll_ids2 = [coll['coll_id'] for coll in colls2]
                coll_ids1.sort()
                coll_ids2.sort()
                if not coll_ids1 == coll_ids2:
                    return False
    else:
        trans1 = merge_dicts(req_trans_colls1.values())
        trans2 = merge_dicts(req_trans_colls2.values())
        tran_ids1 = list(trans1.keys())
        tran_ids2 = list(trans2.keys())
        if not tran_ids1 == tran_ids2:
            return False

        for tran_id in tran_ids1:
            colls1 = trans1[tran_id]
            colls2 = trans2[tran_id]
            coll_ids1 = [coll['coll_id'] for coll in colls1]
            coll_ids2 = [coll['coll_id'] for coll in colls2]
            coll_ids1.sort()
            coll_ids2.sort()
            if not coll_ids1 == coll_ids2:
                return False
    return True


def is_same_req_trans_coll_contents(req_trans_colls1, req_trans_colls2, allow_request_id_None=True):
    req_ids1 = list(req_trans_colls1.keys())
    req_ids2 = list(req_trans_colls2.keys())
    if not allow_request_id_None and not (req_ids1 == req_ids2):
        return False

    if not allow_request_id_None:
        for req_id in req_ids1:
            tran_ids1 = list(req_trans_colls1[req_id].keys())
            tran_ids2 = list(req_trans_colls2[req_id].keys())
            if not tran_ids1 == tran_ids2:
                return False

            for tran_id in tran_ids1:
                coll_contents1 = req_trans_colls1[req_id][tran_id]
                coll_contents2 = req_trans_colls2[req_id][tran_id]
                coll_scope_names1 = [scope_name for scope_name in coll_contents1]
                coll_scope_names2 = [scope_name for scope_name in coll_contents2]
                coll_scope_names1.sort()
                coll_scope_names2.sort()
                if not coll_scope_names1 == coll_scope_names1:
                    return False

                for scope_name in coll_scope_names1:
                    contents1 = req_trans_colls1[req_id][tran_id][scope_name]['contents']
                    contents2 = req_trans_colls2[req_id][tran_id][scope_name]['contents']
                    content_ids1 = [content['content_id'] for content in contents1]
                    content_ids2 = [content['content_id'] for content in contents2]
                    if not content_ids1 == content_ids2:
                        return False
    else:
        if True:
            trans1 = merge_dicts(req_trans_colls1.values())
            trans2 = merge_dicts(req_trans_colls2.values())
            tran_ids1 = list(trans1.keys())
            tran_ids2 = list(trans2.keys())
            if not tran_ids1 == tran_ids2:
                return False

            for tran_id in tran_ids1:
                coll_contents1 = trans1[tran_id]
                coll_contents2 = trans2[tran_id]
                coll_scope_names1 = [scope_name for scope_name in coll_contents1]
                coll_scope_names2 = [scope_name for scope_name in coll_contents2]
                coll_scope_names1.sort()
                coll_scope_names2.sort()
                if not coll_scope_names1 == coll_scope_names1:
                    return False

                for scope_name in coll_scope_names1:
                    contents1 = trans1[tran_id][scope_name]['contents']
                    contents2 = trans2[tran_id][scope_name]['contents']
                    content_ids1 = [content['content_id'] for content in contents1]
                    content_ids2 = [content['content_id'] for content in contents2]
                    if not content_ids1 == content_ids2:
                        return False
    return True



import sys            # noqa F401
import datetime       # noqa F401

from idds.common.utils import json_dumps, setup_logging                 # noqa F401
from idds.common.constants import ContentStatus, ContentType, ContentRelationType, ContentLocking          # noqa F401
from idds.core.requests import get_requests              # noqa F401
from idds.core.messages import retrieve_messages         # noqa F401
from idds.core.transforms import get_transforms, get_transform, update_transform          # noqa F401
from idds.core.workprogress import get_workprogresses    # noqa F401
from idds.core.processings import get_processings        # noqa F401
from idds.core import transforms as core_transforms      # noqa F401
from idds.orm.contents import get_input_contents         # noqa F401
from idds.core.transforms import release_inputs_by_collection, release_inputs_by_collection_old     # noqa F401
from idds.workflowv2.workflow import Workflow            # noqa F401
from idds.workflowv2.work import Work                    # noqa F401


setup_logging(__name__)

tfs = get_transforms()
# tfs = get_transforms(transform_id=350723)
for tf in tfs:
    print(tf['transform_id'])
    work = tf['transform_metadata']['work']
    name = work.get_work_name()
    update_transform(transform_id=tf['transform_id'], parameters={'name': name})
    # break



#!/usr/bin/env python

import re


def match_pattern_file(pattern, lfn):
    pattern1 = "\\$[_a-zA-Z0-9]+"
    pattern2 = "\\$\\{[_a-zA-Z0-9\\/]+\\}"
    while True:
        m = re.search(pattern1, pattern)
        if m:
            pattern = pattern.replace(m.group(0), "*")
        else:
            break
    while True:
        m = re.search(pattern2, pattern)
        if m:
            pattern = pattern.replace(m.group(0), "*")
        else:
            break
    pattern = pattern.replace(".", "\\.")
    pattern = pattern.replace("*", ".*")
    pattern = pattern + "$"
    print(pattern)

    m = re.search(pattern, lfn)
    if m:
        return True
    return False


if __name__ == "__main__":
    pattern = 'user.tmaeno.$JEDITASKID._${SN/P}.results.json'
    lfn = 'user.tmaeno.2321._23.results.json'
    lfn1 = 'user.tmaeno1.2321._23.results.json'
    lfn2 = 'user.tmaeno.2321._23.results1.json'
    lfn3 = 'user.tmaeno.2321._23.results.json1'
    lfn4 = 'user.tmaeno.27529814._000001.results.json'

    ret = match_pattern_file(pattern, lfn)
    print(ret)
    ret = match_pattern_file(pattern, lfn1)
    print(ret)
    ret = match_pattern_file(pattern, lfn2)
    print(ret)
    ret = match_pattern_file(pattern, lfn3)
    print(ret)
    ret = match_pattern_file(pattern, lfn4)
    print(ret)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2020 - 2021


"""
Test workflow.
"""

import unittest2 as unittest
# from nose.tools import assert_equal
from idds.common.utils import setup_logging

from idds.client.client import Client
from idds.common.constants import RequestType, RequestStatus
from idds.common.utils import get_rest_host

from idds.workflowv2.work import Work   # Parameter, WorkStatus
from idds.workflowv2.workflow import Condition, Workflow


setup_logging(__name__)


class TestWorkflow(unittest.TestCase):

    def init(self):
        # init_p = Parameter({'input_dataset': 'data17:data17.test.raw.1'})
        work1 = Work(executable='/bin/hostname', arguments=None, sandbox=None, work_id=1)
        work2 = Work(executable='echo',
                     arguments='--in=IN_DATASET --out=OUT_DATASET',
                     sandbox=None,
                     work_id=2,
                     primary_input_collection={'scope': 'data17', 'name': 'data17.test.raw.1'},
                     output_collections=[{'scope': 'data17', 'name': 'data17.test.work2'}])
        work3 = Work(executable='echo',
                     arguments='--in=IN_DATASET --out=OUT_DATASET',
                     sandbox=None,
                     work_id=3,
                     primary_input_collection={'scope': 'data17', 'name': 'data17.test.work2'},
                     output_collections=[{'scope': 'data17', 'name': 'data17.test.work3'}])

        workflow = Workflow()
        workflow.add_work(work1, initial=True)
        workflow.add_work(work2, initial=True)
        workflow.add_work(work3, initial=False)

        cond = Condition(cond=work2.is_finished, true_work=work3)
        workflow.add_condition(cond)

        return workflow

    def test_workflow(self):
        """ Workflow: Test workflow """
        # init_p = Parameter({'input_dataset': 'data17:data17.test.raw.1'})
        work1 = Work(executable='/bin/hostname', arguments=None, sandbox=None, work_id=1)
        work2 = Work(executable='echo',
                     arguments='--in=IN_DATASET --out=OUT_DATASET',
                     sandbox=None,
                     work_id=2,
                     primary_input_collection={'scope': 'data17', 'name': 'data17.test.raw.1'},
                     output_collections=[{'scope': 'data17', 'name': 'data17.test.work2'}])
        work3 = Work(executable='echo',
                     arguments='--in=IN_DATASET --out=OUT_DATASET',
                     sandbox=None,
                     work_id=3,
                     primary_input_collection={'scope': 'data17', 'name': 'data17.test.work2'},
                     output_collections=[{'scope': 'data17', 'name': 'data17.test.work3'}])

        workflow = Workflow()
        workflow.add_work(work1, initial=True)
        workflow.add_work(work2, initial=True)
        workflow.add_work(work3, initial=False)

        cond = Condition(cond=work2.is_finished, true_work=work3)
        # print(cond.all_works())
        workflow.add_condition(cond)

        # check
        # workflow_str = workflow.serialize()
        # workflow1 = Workflow.deserialize(workflow_str)
        # print(workflow_str)
        # print(workflow1)
        works = workflow.get_current_works()
        # print([str(work) for work in works])
        # print([w.work_id for w in works])
        assert(works == [])

    def test_workflow_request(self):
        workflow = self.init()

        props = {
            'scope': 'workflow',
            'name': workflow.name,
            'requester': 'panda',
            'request_type': RequestType.Workflow,
            'transform_tag': 'workflow',
            'status': RequestStatus.New,
            'priority': 0,
            'lifetime': 30,
            'request_metadata': {'workload_id': '20776840', 'workflow': workflow}
        }

        # print(props)
        host = get_rest_host()
        client = Client(host=host)
        request_id = client.add_request(**props)
        print(request_id)



#!/usr/bin/env python
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0OA
#
# Authors:
# - Wen Guan, <wen.guan@cern.ch>, 2022

import logging
import uuid

from idds.common.constants import Sections
from idds.common.config import config_has_section, config_list_options

from .localeventbusbackend import LocalEventBusBackend


class Singleton(object):
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
            class_._instance._initialized = False
        return class_._instance


class EventBus(Singleton):
    """
    Event Bus
    """

    def __init__(self, logger=None):
        if not self._initialized:
            self._initialized = True

            super(EventBus, self).__init__()
            self._id = str(uuid.uuid4())[:8]
            self.setup_logger(logger)
            self.config_section = Sections.EventBus
            attrs = self.load_attributes()
            if 'backend' in attrs and attrs['backend'] == 'message':
                # ToBeDone
                # self.backend = MsgEventBusBackend(**attrs)
                pass
            else:
                self.backend = LocalEventBusBackend(logger=self.logger, **attrs)

    def setup_logger(self, logger=None):
        """
        Setup logger
        """
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(self.get_class_name())

    def get_class_name(self):
        return self.__class__.__name__

    def load_attributes(self):
        self.logger.info("Loading config for section: %s" % self.config_section)
        attrs = {}
        if config_has_section(self.config_section):
            options = config_list_options(self.config_section)
            for option, value in options:
                if isinstance(value, str) and value.lower() == 'true':
                    value = True
                if isinstance(value, str) and value.lower() == 'false':
                    value = False
                attrs[option] = value
        return attrs

    def publish_event(self, event):
        self.backend.send(event)

    def get_event(self, event_type):
        # demand_event = DemandEvent(event._event_type, self._id)
        event = self.backend.get(event_type, wait=10)
        return event

    def get(self, event_type):
        return self.get_event(event_type)

    def send(self, event):
        return self.publish_event(event)

    def stop(self):
        self.backend.stop()
